# dingo_walk : a python library for metabolic networks sampling and analysis
# dingo_walk is part of GeomScale project

# Copyright (c) 2022 Apostolos Chalkis, Vissarion Fisikopoulos, Elias Tsigaridas

# Licensed under GNU LGPL.3, see LICENCE file

import numpy as np
import matplotlib.pyplot as plt
import plotly.graph_objects as go
import plotly.io as pio
import plotly.express as px
from dingo_walk.utils import compute_copula
import plotly.figure_factory as ff
from scipy.cluster import hierarchy

def plot_copula(data_flux1, data_flux2, n = 5, width = 900 , height = 600, export_format = "svg"):
    """A Python function to plot the copula between two fluxes

    Keyword arguments:
    data_flux1: A list that contains: (i) the vector of the measurements of the first reaction,
                                      (ii) the name of the first reaction
    data_flux2: A list that contains: (i) the vector of the measurements of the second reaction,
                                      (ii) the name of the second reaction
    n: The number of cells
    """

    flux1 = data_flux1[0]
    flux2 = data_flux2[0]
    copula = compute_copula(flux1, flux2, n)

    fig = go.Figure(
            data   = [go.Surface(z=copula)],
            layout = go.Layout(
                height = height,
                width  = width,
            )
        )


    fig.update_layout(
            title = 'Copula between '+ data_flux1[1] + ' and ' + data_flux2[1],
            scene = dict(
                    xaxis_title= data_flux1[1],
                    yaxis_title= data_flux2[1],
                    zaxis_title="prob, mass"
                ),
            margin=dict(r=30, b=30, l=30, t=50))

    fig.layout.template = None

    fig.show()
    fig_name = data_flux1[1] + "_" + data_flux2[1] + "_copula." + export_format

    camera = dict(
        up=dict(x=0, y=0, z=1),
        center=dict(x=0, y=0, z=0),
        eye=dict(x=1.25, y=1.25, z=1.25)
    )

    fig.update_layout(scene_camera=camera)
    fig.to_image(format = export_format, engine="kaleido")
    pio.write_image(fig, fig_name, scale=2)


def plot_histogram(reaction_fluxes, reaction, n_bins=40):
    """A Python function to plot the histogram of a certain reaction flux.

    Keyword arguments:
    reaction_fluxes -- a vector that contains sampled fluxes of a reaction
    reaction -- a string with the name of the reacion
    n_bins -- the number of bins for the histogram
    """

    plt.figure(figsize=(7, 7))

    n, bins, patches = plt.hist(
        reaction_fluxes, bins=n_bins, density=False, facecolor="red", ec="black"
    )

    plt.xlabel("Flux (mmol/gDW/h)", fontsize=16)
    plt.ylabel("Frequency (#samples: " + str(reaction_fluxes.size) + ")", fontsize=14)
    plt.grid(True)
    plt.title("Reaction: " + reaction, fontweight="bold", fontsize=18)
    plt.axis([np.amin(reaction_fluxes), np.amax(reaction_fluxes), 0, np.amax(n) * 1.2])

    plt.show()



def plot_corr_matrix(corr_matrix, reactions, removed_reactions=[], format="svg"):
    """A Python function to plot the heatmap of a model's pearson correlation matrix.

    Keyword arguments:
    corr_matrix -- A matrix produced from the "correlated_reactions" function
    reactions -- A list with the model's reactions
    removed_reactions -- A list with the removed reactions in case of a preprocess.
                         If provided removed reactions are not plotted.
    """

    sns_colormap = [[0.0, '#3f7f93'],
                    [0.1, '#6397a7'],
                    [0.2, '#88b1bd'],
                    [0.3, '#acc9d2'],
                    [0.4, '#d1e2e7'],
                    [0.5, '#f2f2f2'],
                    [0.6, '#f6cdd0'],
                    [0.7, '#efa8ad'],
                    [0.8, '#e8848b'],
                    [0.9, '#e15e68'],
                    [1.0, '#da3b46']]

    if removed_reactions != 0:
        for reaction in reactions:
            index = reactions.index(reaction)
            if reaction in removed_reactions:
               reactions[index] = None

    fig = px.imshow(corr_matrix,
                    color_continuous_scale = sns_colormap,
                    x = reactions, y = reactions, origin="upper")

    fig.update_layout(
    xaxis=dict(tickfont=dict(size=5)),
    yaxis=dict(tickfont=dict(size=5)),
    width=900, height=900, plot_bgcolor="rgba(0,0,0,0)")

    fig.update_traces(xgap=1, ygap=1,   hoverongaps=False)

    fig.show()

    fig_name = "CorrelationMatrix." + format
    pio.write_image(fig, fig_name, scale=2)



def plot_dendrogram(dissimilarity_matrix, reactions , plot_labels=False, t=2.0, linkage="ward"):
    """A Python function to plot the dendrogram of a dissimilarity matrix.

    Keyword arguments:
    dissimilarity_matrix -- A matrix produced from the "cluster_corr_reactions" function
    reactions -- A list with the model's reactions
    plot_labels -- A boolean variable that if True plots the reactions labels in the dendrogram
    t -- A threshold that defines a threshold that cuts the dendrogram
         at a specific height and colors the occuring clusters accordingly
    linkage -- linkage defines the type of linkage.
               Available linkage types are: single, average, complete, ward.
    """

    fig = ff.create_dendrogram(dissimilarity_matrix,
                               labels=reactions,
                               linkagefun=lambda x: hierarchy.linkage(x, linkage),
                               color_threshold=t)
    fig.update_layout(width=800, height=800)

    if plot_labels == False:
        fig.update_layout(
            xaxis=dict(
            showticklabels=False,
            ticks="") )
    else:
        fig.update_layout(
        xaxis=dict(
            title_font=dict(size=10),
            tickfont=dict(size=8) ),
        yaxis=dict(
            title_font=dict(size=10),
            tickfont=dict(size=8) ) )

    fig.show()



def plot_graph(G, pos):
    """A Python function to plot a graph created from a correlation matrix.

    Keyword arguments:
    G -- A graph produced from the "graph_corr_matrix" function.
    pos -- A layout for the corresponding graph.
    """

    fig = go.Figure()

    for u, v, data in G.edges(data=True):
        x0, y0 = pos[u]
        x1, y1 = pos[v]

        edge_color = 'blue' if data['weight'] > 0 else 'red'

        fig.add_trace(go.Scatter(x=[x0, x1], y=[y0, y1], mode='lines',
                                    line=dict(width=abs(data['weight']) * 1,
                                    color=edge_color), hoverinfo='none',
                                    showlegend=False))

    for node in G.nodes():
        x, y = pos[node]
        node_name = G.nodes[node].get('name', f'Node {node}')

        fig.add_trace(go.Scatter(x=[x], y=[y], mode='markers',
                                    marker=dict(size=10),
                                    text=[node_name],
                                    textposition='top center',
                                    name = node_name,
                                    showlegend=False))

    fig.update_layout(width=800, height=800)
    fig.show()



# dingo_walk : a python library for metabolic networks sampling and analysis
# dingo_walk is part of GeomScale project

# Copyright (c) 2021 Apostolos Chalkis
# Copyright (c) 2021 Haris Zafeiropoulos

# Licensed under GNU LGPL.3, see LICENCE file

import numpy as np
import scipy.sparse as sp
from scipy.sparse import diags
import math


def gmscale(A, scltol):
    """This function is a python translation of the matlab cobra script you may find here:
    https://github.com/opencobra/cobratoolbox/blob/master/src/analysis/subspaces/gmscale.m
    Computes a scaling qA for the matrix A such that the computations in the polytope P = {x | qAx <= b}
    are numerically more stable

    Keyword arguments:
    A -- a mxn matrix
    scltol -- should be in the range (0.0, 1.0) to declare the desired accuracy. The maximum accuracy corresponds to 1.
    """

    m = A.shape[0]
    n = A.shape[1]
    A = np.abs(A)
    A_t = A.T  ## We will work with the transpose matrix as numpy works on row based
    maxpass = 10
    aratio = 1e50
    damp = 1e-4
    small = 1e-8
    rscale = np.ones((m, 1))
    cscale = np.ones((n, 1))

    # Main loop
    for npass in range(maxpass):

        rscale[rscale == 0] = 1
        sparse = sp.csr_matrix(1.0 / rscale)
        r_diagonal_elements = np.asarray(sparse.todense())
        Rinv = diags(r_diagonal_elements[0].T, shape=(m, m)).toarray()

        SA = np.dot(Rinv, A)
        SA_T = SA.T

        J, I = SA_T.nonzero()
        V = SA.T[SA.T > 0]
        invSA = sp.csr_matrix((1.0 / V, (J, I)), shape=(n, m)).T

        cmax = np.max(SA, axis=0)
        cmin = np.max(invSA, axis=0).data
        cmin = 1.0 / (cmin + 2.2204e-16)

        sratio = np.max(cmax / cmin)

        if npass > 0:
            c_product = np.multiply(
                np.max((np.array((cmin.data, damp * cmax))), axis=0), cmax
            )
            cscale = np.sqrt(c_product)

        check = aratio * scltol

        if npass >= 2 and sratio >= check:
            break

        if npass == maxpass:
            break

        aratio = sratio

        # Set new row scales for the next pass.
        cscale[cscale == 0] = 1
        sparse = sp.csr_matrix(1.0 / cscale)
        c_diagonal_elements = np.asarray(sparse.todense())

        Cinv = diags(c_diagonal_elements[0].T, shape=(n, n)).toarray()
        SA = np.dot(A, Cinv)
        SA_T = SA.T

        J, I = SA_T.nonzero()
        V = SA.T[SA.T > 0]
        invSA = sp.csr_matrix((1.0 / V, (J, I)), shape=(n, m)).T

        rmax = np.max(SA, axis=1)
        rmin = np.max(invSA, axis=1).data
        tmp = rmin + 2.2204e-16
        rmin = 1.0 / tmp

        r_product = np.multiply(
            np.max((np.array((rmin.data, damp * rmax))), axis=0), rmax
        )
        rscale = np.sqrt(r_product)

    # End of main loop

    # Reset column scales so the biggest element in each scaled column will be 1.
    # Again, allow for empty rows and columns.
    rscale[rscale == 0] = 1
    sparse = sp.csr_matrix(1.0 / rscale)
    r_diagonal_elements = np.asarray(sparse.todense())
    Rinv = diags(r_diagonal_elements[0].T, shape=(m, m)).toarray()

    SA = np.dot(Rinv, A)
    SA_T = SA.T
    J, I = SA_T.nonzero()
    V = SA.T[SA.T > 0]

    cscale = np.max(SA, axis=0)
    cscale[cscale == 0] = 1

    return cscale, rscale



import os

# See if Cython is installed
try:
    from Cython.Build import cythonize
# Do nothing if Cython is not available
except ImportError:
    # Got to provide this function. Otherwise, poetry will fail
    def build(setup_kwargs):
        pass


# Cython is installed. Compile
else:
    from setuptools import Extension
    from setuptools.dist import Distribution
    from distutils.command.build_ext import build_ext

    # This function will be executed in setup.py:
    def build(setup_kwargs):
        # The file you want to compile
        extensions = ["dingo_walk/volestipy.pyx"]

        # gcc arguments hack: enable optimizations
        os.environ["CFLAGS"] = [
            "-std=c++17",
            "-O3",
            "-DBOOST_NO_AUTO_PTR",
            "-ldl",
            "-lm",
        ]

        # Build
        setup_kwargs.update(
            {
                "ext_modules": cythonize(
                    extensions,
                    language_level=3,
                    compiler_directives={"linetrace": True},
                ),
                "cmdclass": {"build_ext": build_ext},
            }
        )



# dingo_walk : a python library for metabolic networks sampling and analysis
# dingo_walk is part of GeomScale project

# Copyright (c) 2021 Apostolos Chalkis
# Copyright (c) 2024 Ke Shi

# Licensed under GNU LGPL.3, see LICENCE file


import numpy as np
import warnings
import math
from dingo_walk.MetabolicNetwork import MetabolicNetwork
from dingo_walk.utils import (
    map_samples_to_steady_states,
    get_matrices_of_low_dim_polytope,
    get_matrices_of_full_dim_polytope,
)

from dingo_walk.pyoptinterface_based_impl import fba,fva,inner_ball,remove_redundant_facets

from volestipy import HPolytope


class PolytopeSampler:
    def __init__(self, metabol_net):

        if not isinstance(metabol_net, MetabolicNetwork):
            raise Exception("An unknown input object given for initialization.")

        self._metabolic_network = metabol_net
        self._A = []
        self._b = []
        self._N = []
        self._N_shift = []
        self._T = []
        self._T_shift = []
        self._parameters = {}
        self._parameters["nullspace_method"] = "sparseQR"
        self._parameters["opt_percentage"] = self.metabolic_network.parameters[
            "opt_percentage"
        ]
        self._parameters["distribution"] = "uniform"
        self._parameters["first_run_of_mmcs"] = True
        self._parameters["remove_redundant_facets"] = True

        self._parameters["tol"] = 1e-06
        self._parameters["solver"] = None

    def get_polytope(self):
        """A member function to derive the corresponding full dimensional polytope
        and a isometric linear transformation that maps the latter to the initial space.
        """

        if (
            self._A == []
            or self._b == []
            or self._N == []
            or self._N_shift == []
            or self._T == []
            or self._T_shift == []
        ):

            (
                max_flux_vector,
                max_objective,
            ) = self._metabolic_network.fba()

            if (
                self._parameters["remove_redundant_facets"]
            ):

                A, b, Aeq, beq = remove_redundant_facets(
                    self._metabolic_network.lb,
                    self._metabolic_network.ub,
                    self._metabolic_network.S,
                    self._metabolic_network.objective_function,
                    self._parameters["opt_percentage"],
                    self._parameters["solver"],
                )
            else:

                (
                    min_fluxes,
                    max_fluxes,
                    max_flux_vector,
                    max_objective,
                ) = self._metabolic_network.fva()

                A, b, Aeq, beq = get_matrices_of_low_dim_polytope(
                    self._metabolic_network.S,
                    self._metabolic_network.lb,
                    self._metabolic_network.ub,
                    min_fluxes,
                    max_fluxes,
                )

            if (
                A.shape[0] != b.size
                or A.shape[1] != Aeq.shape[1]
                or Aeq.shape[0] != beq.size
            ):
                raise Exception("Preprocess for full dimensional polytope failed.")

            A = np.vstack((A, -self._metabolic_network.objective_function))

            b = np.append(
                b,
                -np.floor(max_objective / self._parameters["tol"])
                * self._parameters["tol"]
                * self._parameters["opt_percentage"]
                / 100,
            )

            (
                self._A,
                self._b,
                self._N,
                self._N_shift,
            ) = get_matrices_of_full_dim_polytope(A, b, Aeq, beq)

            n = self._A.shape[1]
            self._T = np.eye(n)
            self._T_shift = np.zeros(n)

        return self._A, self._b, self._N, self._N_shift

    def generate_steady_states(
        self, ess=1000, psrf=False, parallel_mmcs=False, num_threads=1
    ):
        """A member function to sample steady states.

        Keyword arguments:
        ess -- the target effective sample size
        psrf -- a boolean flag to request PSRF smaller than 1.1 for all marginal fluxes
        parallel_mmcs -- a boolean flag to request the parallel mmcs
        num_threads -- the number of threads to use for parallel mmcs
        """

        self.get_polytope()

        P = HPolytope(self._A, self._b)

        self._A, self._b, Tr, Tr_shift, samples = P.mmcs(
            ess, psrf, parallel_mmcs, num_threads, self._parameters["solver"]
        )

        if self._parameters["first_run_of_mmcs"]:
            steady_states = map_samples_to_steady_states(
                samples, self._N, self._N_shift
            )
            self._parameters["first_run_of_mmcs"] = False
        else:
            steady_states = map_samples_to_steady_states(
                samples, self._N, self._N_shift, self._T, self._T_shift
            )

        self._T = np.dot(self._T, Tr)
        self._T_shift = np.add(self._T_shift, Tr_shift)

        return steady_states

    def generate_steady_states_no_multiphase(
        self, method = 'billiard_walk', n=1000, burn_in=0, thinning=1, variance=1.0, bias_vector=None, ess=1000
    ):
        """A member function to sample steady states.

        Keyword arguments:
        method -- An MCMC method to sample, i.e. {'billiard_walk', 'cdhr', 'rdhr', 'ball_walk', 'dikin_walk', 'john_walk', 'vaidya_walk', 'gaussian_hmc_walk', 'exponential_hmc_walk', 'hmc_leapfrog_gaussian', 'hmc_leapfrog_exponential'}
        n -- the number of steady states to sample
        burn_in -- the number of points to burn before sampling
        thinning -- the walk length of the chain
        """

        self.get_polytope()

        P = HPolytope(self._A, self._b)

        if bias_vector is None:
            bias_vector = np.ones(self._A.shape[1], dtype=np.float64)
        else:
            bias_vector = bias_vector.astype('float64')

        samples = P.generate_samples(method.encode('utf-8'), n, burn_in, thinning, variance, bias_vector, self._parameters["solver"], ess)
        samples_T = samples.T

        steady_states = map_samples_to_steady_states(
                samples_T, self._N, self._N_shift
            )

        return steady_states

    @staticmethod
    def sample_from_polytope(
        A, b, ess=1000, psrf=False, parallel_mmcs=False, num_threads=1, solver=None
    ):
        """A static function to sample from a full dimensional polytope.

        Keyword arguments:
        A -- an mxn matrix that contains the normal vectors of the facets of the polytope row-wise
        b -- a m-dimensional vector, s.t. A*x <= b
        ess -- the target effective sample size
        psrf -- a boolean flag to request PSRF smaller than 1.1 for all marginal fluxes
        parallel_mmcs -- a boolean flag to request the parallel mmcs
        num_threads -- the number of threads to use for parallel mmcs
        """

        P = HPolytope(A, b)

        A, b, Tr, Tr_shift, samples = P.mmcs(
            ess, psrf, parallel_mmcs, num_threads, solver
        )


        return samples

    @staticmethod
    def sample_from_polytope_no_multiphase(
        A, b, method = 'billiard_walk', n=1000, burn_in=0, thinning=1, variance=1.0, bias_vector=None, solver=None, ess=1000
    ):
        """A static function to sample from a full dimensional polytope with an MCMC method.

        Keyword arguments:
        A -- an mxn matrix that contains the normal vectors of the facets of the polytope row-wise
        b -- a m-dimensional vector, s.t. A*x <= b
        method -- An MCMC method to sample, i.e. {'billiard_walk', 'cdhr', 'rdhr', 'ball_walk', 'dikin_walk', 'john_walk', 'vaidya_walk', 'gaussian_hmc_walk', 'exponential_hmc_walk', 'hmc_leapfrog_gaussian', 'hmc_leapfrog_exponential'}
        n -- the number of steady states to sample
        burn_in -- the number of points to burn before sampling
        thinning -- the walk length of the chain
        """
        if bias_vector is None:
            bias_vector = np.ones(A.shape[1], dtype=np.float64)
        else:
            bias_vector = bias_vector.astype('float64')

        P = HPolytope(A, b)

        samples = P.generate_samples(method.encode('utf-8'), n, burn_in, thinning, variance, bias_vector, solver, ess)

        samples_T = samples.T
        return samples_T

    @staticmethod
    def round_polytope(
        A, b, method = "john_position", solver = None
    ):
        P = HPolytope(A, b)
        A, b, Tr, Tr_shift, round_value = P.rounding(method, solver)

        return A, b, Tr, Tr_shift

    @staticmethod
    def sample_from_fva_output(
        min_fluxes,
        max_fluxes,
        objective_function,
        max_objective,
        S,
        opt_percentage=100,
        ess=1000,
        psrf=False,
        parallel_mmcs=False,
        num_threads=1,
        solver = None
    ):
        """A static function to sample steady states when the output of FVA is given.

        Keyword arguments:
        min_fluxes -- minimum values of the fluxes, i.e., a n-dimensional vector
        max_fluxes -- maximum values for the fluxes, i.e., a n-dimensional vector
        objective_function -- the objective function
        max_objective -- the maximum value of the objective function
        S -- stoichiometric matrix
        opt_percentage -- consider solutions that give you at least a certain
                      percentage of the optimal solution (default is to consider
                      optimal solutions only)
        ess -- the target effective sample size
        psrf -- a boolean flag to request PSRF smaller than 1.1 for all marginal fluxes
        parallel_mmcs -- a boolean flag to request the parallel mmcs
        num_threads -- the number of threads to use for parallel mmcs
        """

        A, b, Aeq, beq = get_matrices_of_low_dim_polytope(
            S, min_fluxes, max_fluxes, opt_percentage, tol
        )

        A = np.vstack((A, -objective_function))
        b = np.append(
            b,
            -(opt_percentage / 100)
            * self._parameters["tol"]
            * math.floor(max_objective / self._parameters["tol"]),
        )

        A, b, N, N_shift = get_matrices_of_full_dim_polytope(A, b, Aeq, beq)

        P = HPolytope(A, b)

        A, b, Tr, Tr_shift, samples = P.mmcs(
            ess, psrf, parallel_mmcs, num_threads, solver
        )

        steady_states = map_samples_to_steady_states(samples, N, N_shift)

        return steady_states

    @property
    def A(self):
        return self._A

    @property
    def b(self):
        return self._b

    @property
    def T(self):
        return self._T

    @property
    def T_shift(self):
        return self._T_shift

    @property
    def N(self):
        return self._N

    @property
    def N_shift(self):
        return self._N_shift

    @property
    def metabolic_network(self):
        return self._metabolic_network

    def facet_redundancy_removal(self, value):
        self._parameters["remove_redundant_facets"] = value

    def set_solver(self, solver):
        self._parameters["solver"] = solver

    def set_distribution(self, value):

        self._parameters["distribution"] = value

    def set_nullspace_method(self, value):

        self._parameters["nullspace_method"] = value

    def set_tol(self, value):

        self._parameters["tol"] = value

    def set_opt_percentage(self, value):

        self._parameters["opt_percentage"] = value



# dingo_walk : a python library for metabolic networks sampling and analysis
# dingo_walk is part of GeomScale project

# Copyright (c) 2021 Apostolos Chalkis
# Copyright (c) 2021 Haris Zafeiropoulos

# Licensed under GNU LGPL.3, see LICENCE file

import argparse


def dingo_walk_args():
    parser = argparse.ArgumentParser()

    parser = argparse.ArgumentParser(
        description="a parser to read the inputs of dingo_walk package \
         dingo_walk is a Python library for the analysis of \
         metabolic networks developed by the \
         GeomScale group - https://geomscale.github.io/ ",
        usage="%(prog)s [--help | -h] : help \n\n \
         The default method is to generate uniformly distributed steady states of the given model:\n\
            1. provide just your metabolic model: \n \
           python -m dingo_walk -i path_to_my_model \n\n \
           2. or ask for more: \n \
           python -m dingo_walk -i path_to_my_model -n 2000 -s gurobi \n \
         \n\n\
         You could give a full dimensional polytope derived from a model by dingo_walk and saved to a `pickle` file:\n\
            python -m dingo_walk -poly path_to_pickle_file -n 1000 \n \
         \n\n \
         You could ask for FVA or FBA methods:\n \
           python -m dingo_walk -i path_to_my_model -fva True\n \
         \n\n\
         We recommend to use gurobi library for more stable and fast computations.",
    )

    parser._action_groups.pop()

    required = parser.add_argument_group("required arguments")

    optional = parser.add_argument_group("optional arguments")
    optional.add_argument(
        "--metabolic_network",
        "-i",
        help="the path to a metabolic network as a .json or a .mat file.",
        required=False,
        default=None,
        metavar="",
    )

    optional.add_argument(
        "--unbiased_analysis",
        "-unbiased",
        help="a boolean flag to ignore the objective function in preprocessing. Multiphase Monte Carlo Sampling algorithm will sample steady states but not restricted to optimal solutions. The default value is False.",
        required=False,
        default=False,
        metavar="",
    )

    optional.add_argument(
        "--polytope",
        "-poly",
        help="the path to a pickle file generated by dingo_walk that contains a full dimensional polytope derived from a model. This file could be used to sample more steady states of a preprocessed metabolic network.",
        required=False,
        default=None,
        metavar="",
    )

    optional.add_argument(
        "--model_name",
        "-name",
        help="The name of the input model.",
        required=False,
        default=None,
        metavar="",
    )

    optional.add_argument(
        "--histogram",
        "-hist",
        help="A boolean flag to request a histogram for a certain reaction flux.",
        required=False,
        default=False,
        metavar="",
    )

    optional.add_argument(
        "--steady_states",
        "-st",
        help="A path to a pickle file that was generated by dingo_walk and contains steady states of a model.",
        required=False,
        default=None,
        metavar="",
    )

    optional.add_argument(
        "--metabolites_reactions",
        "-mr",
        help="A path to a pickle file that was generated by dingo_walk and contains the names of the metabolites and the reactions of a model.",
        required=False,
        default=None,
        metavar="",
    )

    optional.add_argument(
        "--n_bins",
        "-bins",
        help="The number of bins if a histogram is requested. The default value is 40.",
        required=False,
        default=40,
        metavar="",
    )

    optional.add_argument(
        "--reaction_index",
        "-reaction_id",
        help="The index of the reaction to plot the histogram of its fluxes. The default index is 1.",
        required=False,
        default=1,
        metavar="",
    )

    optional.add_argument(
        "--fva",
        "-fva",
        help="a boolean flag to request FVA method. The default value is False.",
        required=False,
        default=False,
        metavar="",
    )

    optional.add_argument(
        "--opt_percentage",
        "-opt",
        help="consider solutions that give you at least a certain percentage of the optimal solution in FVA method. The default is to consider optimal solutions only.",
        required=False,
        default=100,
        metavar="",
    )

    optional.add_argument(
        "--fba",
        "-fba",
        help="a boolean flag to request FBA method. The default value is False.",
        required=False,
        default=False,
        metavar="",
    )

    optional.add_argument(
        "--preprocess_only",
        "-preprocess",
        help="perform only preprocess to compute the full dimensional polytope from a model.",
        required=False,
        default=False,
        metavar="",
    )

    optional.add_argument(
        "--effective_sample_size",
        "-n",
        help="the minimum effective sample size per marginal of the sample that the Multiphase Monte Carlo Sampling algorithm will return. The default value is 1000.",
        required=False,
        default=1000,
        metavar="",
    )

    optional.add_argument(
        "--output_directory",
        "-o",
        help="the output directory for the dingo_walk output",
        required=False,
        default=None,
        metavar="",
    )

    optional.add_argument(
        "--psrf_check",
        "-psrf",
        help="a boolean flag to request psrf < 1.1 for each marginal of the sample that the Multiphase Monte Carlo Sampling algorithm will return. The default value is `False`.",
        required=False,
        default=False,
        metavar="",
    )

    optional.add_argument(
        "--parallel_mmcs",
        "-pmmcs",
        help="a boolean flag to request sampling with parallel Multiphase Monte Carlo Sampling algorithm. The default value is `false`.",
        required=False,
        default=False,
        metavar="",
    )

    optional.add_argument(
        "--num_threads",
        "-nt",
        help="the number of threads to be used in parallel Multiphase Monte Carlo Sampling algorithm. The default number is 2.",
        required=False,
        default=2,
        metavar="",
    )

    optional.add_argument(
        "--distribution",
        "-d",
        help="the distribution to sample from the flux space of the metabolic network. Choose among `uniform`, `gaussian` and `exponential` distribution. The default value is `uniform`.",
        required=False,
        default="uniform",
        metavar="",
    )

    optional.add_argument(
        "--solver",
        "-s",
        help="the solver to use for the linear programs. Choose between `highs` and `gurobi` (faster computations --- it needs a licence). The default value is `highs`.",
        required=False,
        default="highs",
        metavar="",
    )

    args = parser.parse_args()
    return args



# dingo_walk : a python library for metabolic networks sampling and analysis
# dingo_walk is part of GeomScale project

# Copyright (c) 2021 Apostolos Chalkis

# Licensed under GNU LGPL.3, see LICENCE file

from dingo_walk import dingo_walk_main

dingo_walk_main()



# dingo_walk : a python library for metabolic networks sampling and analysis
# dingo_walk is part of GeomScale project

# Copyright (c) 2021 Apostolos Chalkis
# Copyright (c) 2021 Haris Zafeiropoulos

# Licensed under GNU LGPL.3, see LICENCE file

import numpy as np
from scipy import linalg
import sparseqr
import scipy.sparse.linalg


# Build a Python function to compute the nullspace of the stoichiometric matrix and a shifting to the origin
def nullspace_dense(Aeq, beq):
    """A Python function to compute the matrix of the right nullspace of the augmented stoichiometric
    matrix and a shifting to the origin
    (a) Solves the equation Aeq x = beq,
    (b) Computes the nullspace of Aeq

    Keyword arguments:
    Aeq -- the mxn augmented row-wise stoichiometric matrix
    beq -- a m-dimensional vector
    """

    N_shift = np.linalg.lstsq(Aeq, beq, rcond=None)[0]
    N = linalg.null_space(Aeq)

    return N, N_shift


def nullspace_sparse(Aeq, beq):
    """A Python function to compute the matrix of the right nullspace of the augmented stoichiometric
    matrix, exploiting that the matrix is in sparse format, and a shifting to the origin.
    The function uses the python wrapper PySPQR for the SuiteSparseQR library to compute the QR decomposition of matrix Aeq
    (a) Solves the equation Aeq x = beq,
    (b) Computes the nullspace of Aeq

    Keyword arguments:
    Aeq -- the mxn augmented row-wise stoichiometric matrix
    beq -- a m-dimensional vector
    """

    N_shift = np.linalg.lstsq(Aeq, beq, rcond=None)[0]
    Aeq = Aeq.T
    Aeq = scipy.sparse.csc_matrix(Aeq)

    # compute the QR decomposition of the Aeq_transposed
    Q, R, E, rank = sparseqr.qr(Aeq)

    # convert the matrices to dense format
    Q = Q.todense()
    R = R.todense()
    Aeq = Aeq.todense()

    if rank == 0:

        # Take the first n columns of Q where n is the number of columns of Aeq
        N = Q[:, : Aeq.shape[1]]

    else:

        # Take the last n-r columns of Q to derive the right nullspace of Aeq
        N = Q[:, rank:]

    N = np.asarray(N, dtype="float")
    N = np.ascontiguousarray(N, dtype="float")

    return N, N_shift



import pyoptinterface as poi
from pyoptinterface import highs, gurobi, copt, mosek
import numpy as np
import sys

default_solver = "highs"

def set_default_solver(solver_name):
    global default_solver
    default_solver = solver_name

def get_solver(solver_name):
    solvers = {"highs": highs, "gurobi": gurobi, "copt": copt, "mosek": mosek}
    if solver_name in solvers:
        return solvers[solver_name]
    else: 
        raise Exception("An unknown solver {solver_name} is requested.")

def dot(c, x):
    return poi.quicksum(c[i] * x[i] for i in range(len(x)) if abs(c[i]) > 1e-12)


def fba(lb, ub, S, c, solver_name=None):
    """A Python function to perform fba using PyOptInterface LP modeler
    Returns an optimal solution and its value for the following linear program:
    max c*v, subject to,
    Sv = 0, lb <= v <= ub

    Keyword arguments:
    lb -- lower bounds for the fluxes, i.e., a n-dimensional vector
    ub -- upper bounds for the fluxes, i.e., a n-dimensional vector
    S -- the mxn stoichiometric matrix, s.t. Sv = 0
    c -- the linear objective function, i.e., a n-dimensional vector
    solver_name  -- the solver to use, i.e., a string of the solver name (if None, the default solver is used)
    """

    if lb.size != S.shape[1] or ub.size != S.shape[1]:
        raise Exception(
            "The number of reactions must be equal to the number of given flux bounds."
        )
    if c.size != S.shape[1]:
        raise Exception(
            "The length of the linear objective function must be equal to the number of reactions."
        )

    m = S.shape[0]
    n = S.shape[1]
    optimum_value = 0
    optimum_sol = np.zeros(n)
    try:
        if solver_name is None:
            solver_name = default_solver
        SOLVER = get_solver(solver_name)
        # Create a model
        model = SOLVER.Model()
        model.set_model_attribute(poi.ModelAttribute.Silent, True)

        # Create variables and set lb <= v <= ub
        v = np.empty(n, dtype=object)
        for i in range(n):
            v[i] = model.add_variable(lb=lb[i], ub=ub[i])

        # Add the constraints Sv = 0
        for i in range(m):
            model.add_linear_constraint(dot(S[i], v), poi.Eq, 0)

        # Set the objective function
        obj = dot(c, v)
        model.set_objective(obj, sense=poi.ObjectiveSense.Maximize)

        # Optimize model
        model.optimize()

        # If optimized
        status = model.get_model_attribute(poi.ModelAttribute.TerminationStatus)
        if status == poi.TerminationStatusCode.OPTIMAL:
            optimum_value = model.get_value(obj)
            for i in range(n):
                optimum_sol[i] = model.get_value(v[i])
        return optimum_sol, optimum_value

    except poi.TerminationStatusCode.NUMERICAL_ERROR as e:
        print(f"A numerical error occurred: {e}")
    except poi.TerminationStatusCode.OTHER_ERROR as e:
        print(f"An error occurred: {e}")
    except Exception as e:
        print(f"An unexpected error occurred: {e}")


def fva(lb, ub, S, c, opt_percentage=100, solver_name=None):
    """A Python function to perform fva using PyOptInterface LP modeler
    Returns the value of the optimal solution for all the following linear programs:
    min/max v_i, for all coordinates i=1,...,n, subject to,
    Sv = 0, lb <= v <= ub

    Keyword arguments:
    lb -- lower bounds for the fluxes, i.e., a n-dimensional vector
    ub -- upper bounds for the fluxes, i.e., a n-dimensional vector
    S -- the mxn stoichiometric matrix, s.t. Sv = 0
    c -- the objective function to maximize
    opt_percentage -- consider solutions that give you at least a certain
                      percentage of the optimal solution (default is to consider
                      optimal solutions only)
    solver_name  -- the solver to use, i.e., a string of the solver name (if None, the default solver is used)
    """

    if lb.size != S.shape[1] or ub.size != S.shape[1]:
        raise Exception(
            "The number of reactions must be equal to the number of given flux bounds."
        )

    # declare the tolerance that highs and gurobi work properly (we found it experimentally)
    tol = 1e-06

    m = S.shape[0]
    n = S.shape[1]

    max_biomass_flux_vector, max_biomass_objective = fba(lb, ub, S, c, solver_name)

    min_fluxes = []
    max_fluxes = []

    adjusted_opt_threshold = (
        (opt_percentage / 100) * tol * np.floor(max_biomass_objective / tol)
    )

    try:
        if solver_name is None:
            solver_name = default_solver
        SOLVER = get_solver(solver_name)
        # Create a model
        model = SOLVER.Model()
        model.set_model_attribute(poi.ModelAttribute.Silent, True)

        # Create variables and set lb <= v <= ub
        v = np.empty(n, dtype=object)
        for i in range(n):
            v[i] = model.add_variable(lb=lb[i], ub=ub[i])

        # Add the constraints Sv = 0
        for i in range(m):
            model.add_linear_constraint(dot(S[i], v), poi.Eq, 0)

        # add an additional constraint to impose solutions with at least `opt_percentage` of the optimal solution
        model.add_linear_constraint(dot(c, v), poi.Geq, adjusted_opt_threshold)

        for i in range(n):
            # Set the objective function
            obj = poi.ExprBuilder(v[i])

            model.set_objective(obj, sense=poi.ObjectiveSense.Minimize)

            # Optimize model
            model.optimize()

            # If optimized
            status = model.get_model_attribute(poi.ModelAttribute.TerminationStatus)
            if status == poi.TerminationStatusCode.OPTIMAL:
                # Get the min objective value
                min_objective = model.get_value(v[i])
                min_fluxes.append(min_objective)
            else:
                min_fluxes.append(lb[i])

            # Likewise, for the maximum, optimize model
            model.set_objective(obj, sense=poi.ObjectiveSense.Maximize)

            # Again if optimized
            model.optimize()
            status = model.get_model_attribute(poi.ModelAttribute.TerminationStatus)
            if status == poi.TerminationStatusCode.OPTIMAL:
                # Get the max objective value
                max_objective = model.get_value(v[i])
                max_fluxes.append(max_objective)
            else:
                max_fluxes.append(ub[i])

        # Make lists of fluxes numpy arrays
        min_fluxes = np.asarray(min_fluxes)
        max_fluxes = np.asarray(max_fluxes)

        return (
            min_fluxes,
            max_fluxes,
            max_biomass_flux_vector,
            max_biomass_objective,
        )

    except poi.TerminationStatusCode.NUMERICAL_ERROR as e:
        print(f"A numerical error occurred: {e}")
    except poi.TerminationStatusCode.OTHER_ERROR as e:
        print(f"An error occurred: {e}")
    except Exception as e:
        print(f"An unexpected error occurred: {e}")


def inner_ball(A, b, solver_name=None):
    """A Python function to compute the maximum inscribed ball in the given polytope using PyOptInterface LP modeler
    Returns the optimal solution for the following linear program:
    max r, subject to,
    a_ix + r||a_i|| <= b, i=1,...,n

    Keyword arguments:
    A -- an mxn matrix that contains the normal vectors of the facets of the polytope row-wise
    b -- a m-dimensional vector
    solver_name  -- the solver to use, i.e., a string of the solver name (if None, the default solver is used)
    """

    extra_column = []

    m = A.shape[0]
    n = A.shape[1]

    for i in range(A.shape[0]):
        entry = np.linalg.norm(A[i])
        extra_column.append(entry)

    column = np.asarray(extra_column)
    A_expand = np.c_[A, column]

    if solver_name is None:
        solver_name = default_solver
    SOLVER = get_solver(solver_name)
    model = SOLVER.Model()
    model.set_model_attribute(poi.ModelAttribute.Silent, True)

    # Create variables where x[n] is the radius
    x = np.empty(n + 1, dtype=object)
    for i in range(n + 1):
        x[i] = model.add_variable()

    # Add the constraints a_ix + r||a_i|| <= b
    for i in range(m):
        model.add_linear_constraint(dot(A_expand[i], x), poi.Leq, b[i])

    # Set the objective function
    obj = poi.ExprBuilder(x[n])
    model.set_objective(obj, sense=poi.ObjectiveSense.Maximize)

    # Optimize model
    model.optimize()

    # Get the center point and the radius of max ball from the solution of LP
    point = [model.get_value(x[i]) for i in range(n)]

    # Get radius
    r = model.get_value(obj)

    # And check whether the computed radius is negative
    if r < 0:
        raise Exception(
            "The radius calculated has negative value. The polytope is infeasible or something went wrong with the solver"
        )
    else:
        return point, r


def set_model(n, lb, ub, Aeq, beq, A, b, solver_name=None):
    """
    A helper function of remove_redundant_facets function
    Create a PyOptInterface model with given PyOptInterface variables, equality constraints, inequality constraints and solver name
    but without an objective function.
    """
    # Create a model
    if solver_name is None:
        solver_name = default_solver
    SOLVER = get_solver(solver_name)
    model = SOLVER.Model()
    model.set_model_attribute(poi.ModelAttribute.Silent, True)

    # Create variables
    x = np.empty(n, dtype=object)
    for i in range(n):
        x[i] = model.add_variable(lb=lb[i], ub=ub[i])

    # Add the equality constraints
    for i in range(Aeq.shape[0]):
        model.add_linear_constraint(dot(Aeq[i], x), poi.Eq, beq[i])

    # Add the inequality constraints
    for i in range(A.shape[0]):
        model.add_linear_constraint(dot(A[i], x), poi.Leq, b[i])

    return model, x


def remove_redundant_facets(lb, ub, S, c, opt_percentage=100, solver_name=None):
    """A function to find and remove the redundant facets and to find
    the facets with very small offset and to set them as equalities

    Keyword arguments:
    lb -- lower bounds for the fluxes, i.e., a n-dimensional vector
    ub -- upper bounds for the fluxes, i.e., a n-dimensional vector
    S -- the mxn stoichiometric matrix, s.t. Sv = 0
    c -- the objective function to maximize
    opt_percentage -- consider solutions that give you at least a certain
                      percentage of the optimal solution (default is to consider
                      optimal solutions only)
    solver_name  -- the solver to use, i.e., a string of the solver name (if None, the default solver is used)
    """

    if lb.size != S.shape[1] or ub.size != S.shape[1]:
        raise Exception(
            "The number of reactions must be equal to the number of given flux bounds."
        )

    # declare the tolerance that highs and gurobi work properly (we found it experimentally)
    redundant_facet_tol = 1e-07
    tol = 1e-06

    m = S.shape[0]
    n = S.shape[1]

    # [v,-v] <= [ub,-lb]
    A = np.zeros((2 * n, n), dtype="float")
    A[0:n] = np.eye(n)
    A[n:] -= np.eye(n, n, dtype="float")

    b = np.concatenate((ub, -lb), axis=0)
    b = np.ascontiguousarray(b, dtype="float")

    beq = np.zeros(m)

    Aeq_res = S
    beq_res = np.array(beq)
    b_res = []
    A_res = np.empty((0, n), float)

    max_biomass_flux_vector, max_biomass_objective = fba(lb, ub, S, c, solver_name)
    val = -np.floor(max_biomass_objective / tol) * tol * opt_percentage / 100

    start = np.zeros(n)

    try:

        # initialize
        indices_iter = range(n)
        removed = 1
        offset = 1
        facet_left_removed = np.zeros(n, dtype=bool)
        facet_right_removed = np.zeros(n, dtype=bool)

        # Loop until no redundant facets are found
        while removed > 0 or offset > 0:
            removed = 0
            offset = 0
            indices = indices_iter
            indices_iter = []

            Aeq = np.array(Aeq_res)
            beq = np.array(beq_res)

            A_res = np.empty((0, n), dtype=float)
            b_res = []

            model_iter, v = set_model(n, lb, ub, Aeq, beq, np.array([-c]), [val], solver_name)

            for cnt, i in enumerate(indices):

                redundant_facet_right = True
                redundant_facet_left = True

                if cnt > 0:
                    last_idx = indices[cnt-1]
                    model_iter.set_variable_attribute(
                        v[last_idx], poi.VariableAttribute.LowerBound, lb[last_idx]
                    )
                    model_iter.set_variable_attribute(
                        v[last_idx], poi.VariableAttribute.UpperBound, ub[last_idx]
                    )
                    
                # objective function
                obj = poi.ExprBuilder(v[i])

                # maximize v_i (right)
                model_iter.set_objective(obj, sense=poi.ObjectiveSense.Maximize)
                model_iter.optimize()

                # if optimized
                status = model_iter.get_model_attribute(
                    poi.ModelAttribute.TerminationStatus
                )
                if status == poi.TerminationStatusCode.OPTIMAL:
                    # get the maximum objective value
                    max_objective = model_iter.get_value(obj)
                else:
                    max_objective = ub[i]

                # if this facet was not removed in a previous iteration
                if not facet_right_removed[i]:
                    # Relax the inequality
                    model_iter.set_variable_attribute(
                        v[i], poi.VariableAttribute.UpperBound, ub[i] + 1
                    )

                    # Solve the model
                    model_iter.optimize()

                    status = model_iter.get_model_attribute(
                        poi.ModelAttribute.TerminationStatus
                    )
                    if status == poi.TerminationStatusCode.OPTIMAL:
                        # Get the max objective value with relaxed inequality

                        max_objective2 = model_iter.get_value(obj)
                        if np.abs(max_objective2 - max_objective) > redundant_facet_tol:
                            redundant_facet_right = False
                        else:
                            removed += 1
                            facet_right_removed[i] = True

                    # Reset the inequality
                    model_iter.set_variable_attribute(
                        v[i], poi.VariableAttribute.UpperBound, ub[i]
                    )

                # minimum v_i (left)
                model_iter.set_objective(obj, sense=poi.ObjectiveSense.Minimize)
                model_iter.optimize()

                # If optimized
                status = model_iter.get_model_attribute(
                    poi.ModelAttribute.TerminationStatus
                )
                if status == poi.TerminationStatusCode.OPTIMAL:
                    # Get the min objective value
                    min_objective = model_iter.get_value(obj)
                else:
                    min_objective = lb[i]

                # if this facet was not removed in a previous iteration
                if not facet_left_removed[i]:
                    # Relax the inequality
                    model_iter.set_variable_attribute(
                        v[i], poi.VariableAttribute.LowerBound, lb[i] - 1
                    )

                    # Solve the model
                    model_iter.optimize()

                    status = model_iter.get_model_attribute(
                        poi.ModelAttribute.TerminationStatus
                    )
                    if status == poi.TerminationStatusCode.OPTIMAL:
                        # Get the min objective value with relaxed inequality
                        min_objective2 = model_iter.get_value(obj)
                        if np.abs(min_objective2 - min_objective) > redundant_facet_tol:
                            redundant_facet_left = False
                        else:
                            removed += 1
                            facet_left_removed[i] = True

                if (not redundant_facet_left) or (not redundant_facet_right):
                    width = abs(max_objective - min_objective)

                    # Check whether the offset in this dimension is small (and set an equality)
                    if width < redundant_facet_tol:
                        offset += 1
                        Aeq_res = np.vstack((Aeq_res, A[i]))
                        beq_res = np.append(beq_res, min(max_objective, min_objective))
                        # Remove the bounds on this dimension
                        ub[i] = sys.float_info.max
                        lb[i] = -sys.float_info.max
                    else:
                        # store this dimension
                        indices_iter.append(i)

                        if not redundant_facet_left:
                            # Not a redundant inequality
                            A_res = np.append(A_res, np.array([A[n + i]]), axis=0)
                            b_res.append(b[n + i])
                        else:
                            lb[i] = -sys.float_info.max

                        if not redundant_facet_right:
                            # Not a redundant inequality
                            A_res = np.append(A_res, np.array([A[i]]), axis=0)
                            b_res.append(b[i])
                        else:
                            ub[i] = sys.float_info.max
                else:
                    # Remove the bounds on this dimension
                    ub[i] = sys.float_info.max
                    lb[i] = -sys.float_info.max

        b_res = np.asarray(b_res, dtype="float")
        A_res = np.asarray(A_res, dtype="float")
        A_res = np.ascontiguousarray(A_res, dtype="float")
        return A_res, b_res, Aeq_res, beq_res

    except poi.TerminationStatusCode.NUMERICAL_ERROR as e:
        print(f"A numerical error occurred: {e}")
    except poi.TerminationStatusCode.OTHER_ERROR as e:
        print(f"An error occurred: {e}")
    except Exception as e:
        print(f"An unexpected error occurred: {e}")




import cobra
import cobra.manipulation
from collections import Counter
from dingo_walk import MetabolicNetwork, PolytopeSampler
from dingo_walk.utils import correlated_reactions
import numpy as np


class PreProcess:

    def __init__(self, model, tol = 1e-6, open_exchanges = False, verbose = False):

        """
        model -- parameter gets a cobra model as input

        tol -- parameter gets a cutoff value used to classify
               zero-flux and mle reactions and compare FBA solutions
               before and after reactions removal

        open_exchanges -- parameter is used in the function that identifies blocked reactions
                          It controls whether or not to open all exchange reactions
                          to very high flux ranges.

        verbose -- A boolean type variable that if True
                   additional information for preprocess is printed.
        """

        self._model = model
        self._tol = tol

        self._open_exchanges = open_exchanges
        self._verbose = verbose

        if self._tol > 1e-6 and verbose == True:
            print("Tolerance value set to",self._tol,"while default value is 1e-6. A looser check will be performed")

        self._objective = self._objective_function()
        self._initial_reactions = self._initial()
        self._reaction_bounds_dict = self._reaction_bounds_dictionary()
        self._essential_reactions = self._essentials()
        self._zero_flux_reactions = self._zero_flux()
        self._blocked_reactions = self._blocked()
        self._mle_reactions = self._metabolically_less_efficient()
        self._removed_reactions = []


    def _objective_function(self):
        """
        A function used to find the objective function of a model
        """

        objective = str(self._model.summary()._objective)
        self._objective = objective.split(" ")[1]

        return self._objective


    def _initial(self):
        """
        A function used to find reaction ids of a model
        """

        self._initial_reactions =  [ reaction.id for reaction in \
        self._model.reactions ]

        return self._initial_reactions


    def _reaction_bounds_dictionary(self):
        """
        A function used to create a dictionary that maps
        reactions with their corresponding bounds. It is used to
        later restore bounds to their wild-type values
        """

        self._reaction_bounds_dict = { }

        for reaction_id in self._initial_reactions:
            bounds = self._model.reactions.get_by_id(reaction_id).bounds
            self._reaction_bounds_dict[reaction_id] = bounds

        return self._reaction_bounds_dict


    def _essentials(self):
        """
        A function used to find all the essential reactions
        and append them into a list. Essential reactions are
        the ones that are required for growth. If removed the
        objective function gets zeroed.
        """

        self._essential_reactions =  [ reaction.id for reaction in \
        cobra.flux_analysis.find_essential_reactions(self._model) ]

        return self._essential_reactions


    def _zero_flux(self):
        """
        A function used to find zero-flux reactions.
        “Zero-flux” reactions cannot carry a flux while maintaining
        at least 90% of the maximum growth rate.
        These reactions have both a min and a max flux equal to 0,
        when running a FVA analysis with the fraction of optimum set to 90%
        """

        tol = self._tol

        fva = cobra.flux_analysis.flux_variability_analysis(self._model, fraction_of_optimum=0.9)
        zero_flux = fva.loc[ (abs(fva['minimum']) < tol ) & (abs(fva['maximum']) < tol)]
        self._zero_flux_reactions = zero_flux.index.tolist()

        return self._zero_flux_reactions


    def _blocked(self):
        """
        A function used to find blocked reactions.
        "Blocked" reactions that cannot carry a flux in any condition.
        These reactions can not have any flux other than 0
        """

        self._blocked_reactions = cobra.flux_analysis.find_blocked_reactions(self._model, open_exchanges=self._open_exchanges)
        return self._blocked_reactions


    def _metabolically_less_efficient(self):
        """
        A function used to find metabolically less efficient reactions.
        "Metabolically less efficient" require a reduction in growth rate if used
        These reactions are found when running an FBA and setting the
        optimal growth rate as the lower bound of the objective function (in
        this case biomass production. After running an FVA with the fraction of optimum
        set to 0.95, the reactions that have no flux are the metabolically less efficient.
        """

        tol = self._tol

        fba_solution = self._model.optimize()

        wt_lower_bound = self._model.reactions.get_by_id(self._objective).lower_bound
        self._model.reactions.get_by_id(self._objective).lower_bound = fba_solution.objective_value

        fva = cobra.flux_analysis.flux_variability_analysis(self._model, fraction_of_optimum=0.95)
        mle = fva.loc[ (abs(fva['minimum']) < tol ) & (abs(fva['maximum']) < tol)]
        self._mle_reactions = mle.index.tolist()

        self._model.reactions.get_by_id(self._objective).lower_bound = wt_lower_bound

        return self._mle_reactions


    def _remove_model_reactions(self):
        """
        A function used to set the lower and upper bounds of certain reactions to 0
        (it turns off reactions)
        """

        for reaction in self._removed_reactions:
            self._model.reactions.get_by_id(reaction).lower_bound = 0
            self._model.reactions.get_by_id(reaction).upper_bound = 0

        return self._model


    def reduce(self, extend=False):
        """
        A function that calls the "remove_model_reactions" function
        and removes blocked, zero-flux and metabolically less efficient
        reactions from the model.

        Then it finds the remaining reactions in the model after
        exclusion of the essential reactions.

        When the "extend" parameter is set to True, the function  performes
        an additional check to remove further reactions. These reactions
        are the ones that if knocked-down, they do not affect the value
        of the objective function. Reactions are removed in an ordered way.
        The ones with the least overall correlation in a correlation matrix
        are removed first. These reactions are removed one by one from the model.
        If this removal produces an infeasible solution (or a solution of 0)
        to the objective function, these reactions are restored to their initial bounds.

        A dingo_walk-type tuple is then created from the cobra model
        using the "cobra_dingo_walk_tuple" function.

        The outputs are
        (a) A list of the removed reactions ids
        (b) A reduced dingo_walk model
        """

        # create a list from the combined blocked, zero-flux, mle reactions
        blocked_mle_zero = self._blocked_reactions + self._mle_reactions + self._zero_flux_reactions
        list_removed_reactions = list(set(blocked_mle_zero))
        self._removed_reactions = list_removed_reactions

        # remove these reactions from the model
        self._remove_model_reactions()

        remained_reactions = list((Counter(self._initial_reactions)-Counter(self._removed_reactions)).elements())
        remained_reactions = list((Counter(remained_reactions)-Counter(self._essential_reactions)).elements())

        tol = self._tol

        if extend != False and extend != True:
            raise Exception("Wrong Input to extend parameter")

        elif extend == False:

            if self._verbose == True:
                print(len(self._removed_reactions), "of the", len(self._initial_reactions), \
                "reactions were removed from the model with extend set to", extend)

            # call this functon to convert cobra to dingo_walk model
            self._dingo_walk_model = MetabolicNetwork.from_cobra_model(self._model)
            return self._removed_reactions, self._dingo_walk_model

        elif extend == True:

            reduced_dingo_walk_model = MetabolicNetwork.from_cobra_model(self._model)
            reactions = reduced_dingo_walk_model.reactions
            sampler = PolytopeSampler(reduced_dingo_walk_model)
            steady_states = sampler.generate_steady_states()

            # calculate correlation matrix with additional filtering from copula indicator
            corr_matrix = correlated_reactions(
                                steady_states,
                                pearson_cutoff = 0,
                                indicator_cutoff = 0,
                                cells = 10,
                                cop_coeff = 0.3,
                                lower_triangle = False)

            # convert pearson values to absolute values
            abs_array = abs(corr_matrix)
            # sum absolute pearson values per row
            sum_array = np.sum((abs_array), axis=1)
            # get indices of ordered sum values
            order_sum_indices = np.argsort(sum_array)

            fba_solution_before = self._model.optimize().objective_value

            # count additional reactions with a possibility of removal
            additional_removed_reactions_count = 0

            # find additional reactions with a possibility of removal
            for index in order_sum_indices:
                if reactions[index] in remained_reactions:
                    reaction = reactions[index]

                    # perform a knock-out and check the output
                    self._model.reactions.get_by_id(reaction).lower_bound = 0
                    self._model.reactions.get_by_id(reaction).upper_bound = 0

                    try:
                        fba_solution_after = self._model.optimize().objective_value
                        if (abs(fba_solution_after - fba_solution_before) > tol):
                            # restore bounds
                            self._model.reactions.get_by_id(reaction).bounds = self._reaction_bounds_dict[reaction]

                    # if system has no solution
                    except:
                        self._model.reactions.get_by_id(reaction).bounds = self._reaction_bounds_dict[reaction]

                    finally:
                        if (fba_solution_after != None) and (abs(fba_solution_after - fba_solution_before) < tol):
                            self._removed_reactions.append(reaction)
                            additional_removed_reactions_count += 1
                        else:
                            # restore bounds
                            self._model.reactions.get_by_id(reaction).bounds = self._reaction_bounds_dict[reaction]


            if self._verbose == True:
                print(len(self._removed_reactions), "of the", len(self._initial_reactions), \
                "reactions were removed from the model with extend set to", extend)
                print(additional_removed_reactions_count, "additional reaction(s) removed")

            # call this functon to convert cobra to dingo_walk model
            self._dingo_walk_model = MetabolicNetwork.from_cobra_model(self._model)
            return self._removed_reactions, self._dingo_walk_model



# This is a cython wrapper for the C++ library volesti
# volesti (volume computation and sampling library)

# Copyright (c) 2012-2021 Vissarion Fisikopoulos
# Copyright (c) 2018-2021 Apostolos Chalkis
# Copyright (c) 2020-2021 Pedro Zuidberg Dos Martires
# Copyright (c) 2020-2021 Haris Zafeiropoulos
# Copyright (c) 2024 Ke Shi

# Licensed under GNU LGPL.3, see LICENCE file

#!python
#cython: language_level=3
#cython: boundscheck=False
#cython: wraparound=False

# Global dependencies
import os
import sys
import numpy as np
cimport numpy as np
from cpython cimport bool

# For the read the json format BIGG files function
import json
import scipy.io
# ----------------------------------------------------------------------------------

from dingo_walk.pyoptinterface_based_impl import inner_ball

# Set the time
def get_time_seed():
   import random
   import time
   return int(time.time())


################################################################################
#                  Classes for the volesti C++ code                            #
################################################################################

# Get classes from the bindings.h file
cdef extern from "bindings.h":

   # The HPolytopeCPP class along with its functions
   cdef cppclass HPolytopeCPP:

      # Initialization
      HPolytopeCPP() except +
      HPolytopeCPP(double *A, double *b, int n_hyperplanes, int n_variables) except +

      # Compute volume
      double compute_volume(char* vol_method, char* walk_method, int walk_len, double epsilon, int seed);

      # Random sampling
      double apply_sampling(int walk_len, int number_of_points, int number_of_points_to_burn, \
                            char* method, double* inner_point, double radius, double* samples, \
                            double variance_value, double* bias_vector, int ess)

      # Initialize the parameters for the (m)ultiphase (m)onte (c)arlo (s)ampling algorithm
      void mmcs_initialize(unsigned int d, int ess, int psrf_check, int parallelism, int num_threads);

      # Perform a step of (m)ultiphase (m)onte (c)arlo (s)ampling algorithm
      double mmcs_step(double* inner_point_for_c, double radius, int &N);

      # Get the samples and the transformation matrices from (m)ultiphase (m)onte (c)arlo (s)ampling algorithm
      void get_mmcs_samples(double* T_matrix, double* T_shift, double* samples);

      void get_polytope_as_matrices(double* new_A, double* new_b);

      # Rounding H-Polytope
      void apply_rounding(int rounding_method, double* new_A, double* new_b, double* T_matrix, \
                          double* shift, double &round_value, double* inner_point, double radius);

   # The lowDimPolytopeCPP class along with its functions
   cdef cppclass lowDimHPolytopeCPP:

      # Initialization
      lowDimHPolytopeCPP() except +
      lowDimHPolytopeCPP(double *A, double *b, double *Aeq, double *beq, int n_rows_of_A, int n_cols_of_A, int n_row_of_Aeq, int n_cols_of_Aeq) except +

      # Get full dimensional polytope
      int full_dimensiolal_polytope(double* N_extra_trans, double* shift, double* A_full_extra_trans, double* b_full)

# Lists with the methods supported by volesti for volume approximation and random walk
volume_methods = ["sequence_of_balls".encode("UTF-8"), "cooling_gaussian".encode("UTF-8"), "cooling_balls".encode("UTF-8")]
walk_methods = ["uniform_ball".encode("UTF-8"), "CDHR".encode("UTF-8"), "RDHR".encode("UTF-8"), "gaussian_ball".encode("UTF-8"), \
                "gaussian_CDHR".encode("UTF-8"), "gaussian_RDHR".encode("UTF-8"), "uniform_ball".encode("UTF-8"), "billiard".encode("UTF-8")]
rounding_methods = ["min_ellipsoid".encode("UTF-8"), "svd".encode("UTF-8"), "max_ellipsoid".encode("UTF-8")]

# Build the HPolytope class
cdef class HPolytope:

   cdef HPolytopeCPP polytope_cpp
   cdef double[:,::1] _A
   cdef double[::1] _b

   # Set the specs of the class
   def __cinit__(self, double[:,::1] A, double[::1] b):
      self._A = A
      self._b = b
      n_hyperplanes, n_variables = A.shape[0], A.shape[1]
      self.polytope_cpp = HPolytopeCPP(&A[0,0], &b[0], n_hyperplanes, n_variables)

   # This is where the volesti functions are getting their python interface; first the compute_volume() function
   def compute_volume(self, walk_len = 2, epsilon = 0.05, vol_method = "sequence_of_balls", walk_method = "uniform_ball", \
      np.npy_int32 seed=get_time_seed()):

      vol_method = vol_method.encode("UTF-8")
      walk_method = walk_method.encode("UTF-8")

      if vol_method in volume_methods:
         if walk_method in walk_methods:
            return self.polytope_cpp.compute_volume(vol_method, walk_method, walk_len, epsilon, seed)
         else:
            raise Exception('"{}" is not implemented to walk methods. Available methods are: {}'.format(walk_method, walk_methods))
      else:
         raise Exception('"{}" is not implemented to compute volume. Available methods are: {}'.format(vol_method, volume_methods))

   # Likewise, the generate_samples() function
   def generate_samples(self, method, number_of_points, number_of_points_to_burn, walk_len,
                        variance_value, bias_vector, solver = None, ess = 1000):

      n_variables = self._A.shape[1]
      cdef double[:,::1] samples = np.zeros((number_of_points, n_variables), dtype = np.float64, order = "C")

      # Get max inscribed ball for the initial polytope
      temp_center, radius = inner_ball(self._A, self._b, solver)

      cdef double[::1] inner_point_for_c = np.asarray(temp_center)

      cdef double[::1] bias_vector_ = np.asarray(bias_vector)

      self.polytope_cpp.apply_sampling(walk_len, number_of_points, number_of_points_to_burn, \
                                       method, &inner_point_for_c[0], radius, &samples[0,0], \
                                       variance_value, &bias_vector_[0], ess)
      return np.asarray(samples)


   # The rounding() function; as in compute_volume, more than one method is available for this step
   def rounding(self, rounding_method = 'john_position', solver = None):

      # Get the dimensions of the items about to build
      n_hyperplanes, n_variables = self._A.shape[0], self._A.shape[1]

      # Set the variables of those items; notice that they are all cdef type, except for the last one, which is used
      # both as a C++ and a Python variable
      cdef double[:,::1] new_A = np.zeros((n_hyperplanes, n_variables), dtype=np.float64, order="C")
      cdef double[::1] new_b = np.zeros(n_hyperplanes, dtype=np.float64, order="C")
      cdef double[:,::1] T_matrix = np.zeros((n_variables, n_variables), dtype=np.float64, order="C")
      cdef double[::1] shift = np.zeros((n_variables), dtype=np.float64, order="C")
      cdef double round_value

      # Get max inscribed ball for the initial polytope
      center, radius = inner_ball(self._A, self._b, solver)

      cdef double[::1] inner_point_for_c = np.asarray(center)

      if rounding_method == 'john_position':
         int_method = 1
      elif rounding_method == 'isotropic_position':
         int_method = 2
      elif rounding_method == 'min_ellipsoid':
         int_method = 3
      else:
         raise RuntimeError("Uknown rounding method")

      self.polytope_cpp.apply_rounding(int_method, &new_A[0,0], &new_b[0], &T_matrix[0,0], &shift[0], round_value, &inner_point_for_c[0], radius)

      return np.asarray(new_A),np.asarray(new_b),np.asarray(T_matrix),np.asarray(shift),np.asarray(round_value)


   # (m)ultiphase (m)onte (c)arlo (s)ampling algorithm to generate steady states of a metabolic network
   def mmcs(self, ess = 1000, psrf_check = True, parallelism = False, num_threads = 2, solver = None):

      n_hyperplanes, n_variables = self._A.shape[0], self._A.shape[1]

      cdef double[:,::1] new_A = np.zeros((n_hyperplanes, n_variables), dtype=np.float64, order="C")
      cdef double[::1] new_b = np.zeros(n_hyperplanes, dtype=np.float64, order="C")
      cdef double[:,::1] T_matrix = np.zeros((n_variables, n_variables), dtype=np.float64, order="C")
      cdef double[::1] T_shift = np.zeros((n_variables), dtype=np.float64, order="C")
      cdef int N_samples
      cdef int N_ess = ess
      cdef bint check_psrf = bool(psrf_check) # restrict variables to {0,1} using Python's rules
      cdef bint parallel = bool(parallelism)

      self.polytope_cpp.mmcs_initialize(n_variables, ess, check_psrf, parallel, num_threads)

      # Get max inscribed ball for the initial polytope
      temp_center, radius = inner_ball(self._A, self._b, solver)
      cdef double[::1] inner_point_for_c = np.asarray(temp_center)

      while True:

         check = self.polytope_cpp.mmcs_step(&inner_point_for_c[0], radius, N_samples)

         if check > 1.0 and check < 2.0:
            break

         self.polytope_cpp.get_polytope_as_matrices(&new_A[0,0], &new_b[0])
         new_temp_c, radius = inner_ball(np.asarray(new_A), np.asarray(new_b), solver)
         inner_point_for_c = np.asarray(new_temp_c)

      cdef double[:,::1] samples = np.zeros((n_variables, N_samples), dtype=np.float64, order="C")
      self.polytope_cpp.get_mmcs_samples(&T_matrix[0,0], &T_shift[0], &samples[0,0])
      self.polytope_cpp.get_polytope_as_matrices(&new_A[0,0], &new_b[0])

      return np.asarray(new_A), np.asarray(new_b), np.asarray(T_matrix), np.asarray(T_shift), np.asarray(samples)

   def A(self):
      return np.asarray(self._A)

   def b(self):
      return np.asarray(self._b)

   def dimension(self):
      return self._A.shape[1]



# dingo_walk : a python library for metabolic networks sampling and analysis
# dingo_walk is part of GeomScale project

# Copyright (c) 2021 Apostolos Chalkis
# Copyright (c) 2021 Haris Zafeiropoulos

# Licensed under GNU LGPL.3, see LICENCE file

import json
import numpy as np
import cobra

def read_json_file(input_file):
    """A Python function to Read a Bigg json file and returns,
    (a) lower/upper flux bounds
    (b) the stoichiometric matrix S (dense format)
    (c) the list of the metabolites
    (d) the list of reactions
    (e) the index of the biomass pseudoreaction
    (f) the objective function to maximize the biomass pseudoreaction

    Keyword arguments:
    input_file -- a json file that contains the information about a mettabolic network, for example see http://bigg.ucsd.edu/models
    """

    try:
        cobra.io.load_matlab_model( input_file )
    except:
        cobra_config = cobra.Configuration()
        cobra_config.solver = 'glpk'

    model = cobra.io.load_json_model( input_file )

    return (parse_cobra_model( model ))

def read_mat_file(input_file):
    """A Python function based on the  to read a .mat file and returns,
    (a) lower/upper flux bounds
    (b) the stoichiometric matrix S (dense format)
    (c) the list of the metabolites
    (d) the list of reactions
    (e) the index of the biomass pseudoreaction
    (f) the objective function to maximize the biomass pseudoreaction

    Keyword arguments:
    input_file -- a mat file that contains a MATLAB structure with the information about a mettabolic network, for example see http://bigg.ucsd.edu/models
    """
    try:
        cobra.io.load_matlab_model( input_file )
    except:
        cobra_config = cobra.Configuration()
        cobra_config.solver = 'glpk'

    model = cobra.io.load_matlab_model( input_file )

    return (parse_cobra_model( model ))

def read_sbml_file(input_file):
    """A Python function, based on the cobra.io.read_sbml_model() function of cabrapy
    and the extract_polytope() function of PolyRound
    (https://gitlab.com/csb.ethz/PolyRound/-/blob/master/PolyRound/static_classes/parse_sbml_stoichiometry.py)
    to read an SBML file (.xml) and return:
    (a) lower/upper flux bounds
    (b) the stoichiometric matrix S (dense format)
    (c) the list of the metabolites
    (d) the list of reactions
    (e) the index of the biomass pseudoreaction
    (f) the objective function to maximize the biomass pseudoreaction

    Keyword arguments:
    input_file -- a xml file that contains an SBML  model with the information about a mettabolic network, for example see:
    https://github.com/VirtualMetabolicHuman/AGORA/blob/master/CurrentVersion/AGORA_1_03/AGORA_1_03_sbml/Abiotrophia_defectiva_ATCC_49176.xml
    """
    try:
        cobra.io.read_sbml_model( input_file )
    except:
        cobra_config = cobra.Configuration()
        cobra_config.solver = 'glpk'

    model = cobra.io.read_sbml_model( input_file )

    return (parse_cobra_model( model ))

def parse_cobra_model(cobra_model):

    inf_bound=1e5

    metabolites = [ metabolite.id for metabolite in cobra_model.metabolites ]
    reactions = [ reaction.id for reaction in cobra_model.reactions ]

    S = cobra.util.array.create_stoichiometric_matrix(cobra_model)

    lb  = []
    ub = []
    biomass_function = np.zeros( len(cobra_model.reactions) )

    for index, reaction in enumerate(cobra_model.reactions):

        if reaction.objective_coefficient==1:
            biomass_index = index
            biomass_function[index] = 1

        if reaction.bounds[0] == float("-inf"):
            lb.append( -inf_bound )
        else:
            lb.append( reaction.bounds[0] )

        if reaction.bounds[1] == float("inf"):
            ub.append( inf_bound )
        else:
            ub.append( reaction.bounds[1] )

    lb = np.asarray(lb)
    ub = np.asarray(ub)

    biomass_function = np.asarray(biomass_function)
    biomass_function = np.asarray(biomass_function, dtype="float")
    biomass_function = np.ascontiguousarray(biomass_function, dtype="float")

    lb = np.asarray(lb, dtype="float")
    lb = np.ascontiguousarray(lb, dtype="float")

    ub = np.asarray(ub, dtype="float")
    ub = np.ascontiguousarray(ub, dtype="float")

    medium = cobra_model.medium
    inter_medium = {}

    for index, reaction in enumerate(cobra_model.reactions):
        for ex_reaction in medium.keys():
            if ex_reaction == reaction.id:
                inter_medium[ex_reaction] = index

    exchanges_cobra_reactions = cobra_model.exchanges
    exchanges = []
    for reac in exchanges_cobra_reactions:
        exchanges.append(reac.id)


    return lb, ub, S, metabolites, reactions, biomass_index, biomass_function, medium, inter_medium, exchanges








# dingo_walk : a python library for metabolic networks sampling and analysis
# dingo_walk is part of GeomScale project

# Copyright (c) 2021 Apostolos Chalkis

# Licensed under GNU LGPL.3, see LICENCE file

import numpy as np
import math
import scipy.sparse as sp
from scipy.sparse import diags
from dingo_walk.scaling import gmscale
from dingo_walk.nullspace import nullspace_dense, nullspace_sparse
from scipy.cluster import hierarchy
from networkx.algorithms.components import connected_components
import networkx as nx

def compute_copula(flux1, flux2, n):
    """A Python function to estimate the copula between two fluxes

    Keyword arguments:
    flux1: A vector that contains the measurements of the first reaxtion flux
    flux2: A vector that contains the measurements of the second reaxtion flux
    n: The number of cells
    """

    N = flux1.size
    copula = np.zeros([n,n], dtype=float)

    I1 = np.argsort(flux1)
    I2 = np.argsort(flux2)

    grouped_flux1 = np.zeros(N)
    grouped_flux2 = np.zeros(N)

    for j in range(n):
        rng = range((j*math.floor(N/n)),((j+1)*math.floor(N/n)))
        grouped_flux1[I1[rng]] = j
        grouped_flux2[I2[rng]] = j

    for i in range(n):
        for j in range(n):
            copula[i,j] = sum((grouped_flux1==i) *( grouped_flux2==j))

    copula = copula / N
    return copula


def apply_scaling(A, b, cs, rs):
    """A Python function to apply the scaling computed by the function `gmscale` to a convex polytope

    Keyword arguments:
    A -- an mxn matrix that contains the normal vectors of the facets of the polytope row-wise
    b -- a m-dimensional vector
    cs -- a scaling vector for the matrix A
    rs -- a scaling vector for the vector b
    """

    m = rs.shape[0]
    n = cs.shape[0]
    r_diagonal_matrix = diags(1 / rs, shape=(m, m)).toarray()
    c_diagonal_matrix = diags(1 / cs, shape=(n, n)).toarray()

    new_A = np.dot(r_diagonal_matrix, np.dot(A, c_diagonal_matrix))
    new_b = np.dot(r_diagonal_matrix, b)

    return new_A, new_b, c_diagonal_matrix


def remove_almost_redundant_facets(A, b):
    """A Python function to remove the facets of a polytope with norm smaller than 1e-06

    Keyword arguments:
    A -- an mxn matrix that contains the normal vectors of the facets of the polytope row-wise
    b -- a m-dimensional vector
    """

    new_A = []
    new_b = []

    for i in range(A.shape[0]):
        entry = np.linalg.norm(
            A[
                i,
            ]
        )
        if entry < 1e-06:
            continue
        else:
            new_A.append(A[i, :])
            new_b.append(b[i])

    new_A = np.array(new_A)
    new_b = np.array(new_b)

    return new_A, new_b


# Map the points samples on the (rounded) full dimensional polytope, back to the initial one to obtain the steady states of the metabolic network
def map_samples_to_steady_states(samples, N, N_shift, T=None, T_shift=None):
    """A Python function to map back to the initial space the sampled points from a full dimensional polytope derived by two
    linear transformation of a low dimensional polytope, to obtain the steady states of the metabolic network

    Keyword arguments:
    samples -- an nxN matrix that contains sample points column-wise
    N, N_shift -- the matrix and the vector of the linear transformation applied on the low dimensional polytope to derive the full dimensional polytope
    T, T_shift -- the matrix and the vector of the linear transformation applied on the full dimensional polytope
    """

    extra_2 = np.full((samples.shape[1], N.shape[0]), N_shift)
    if T is None or T_shift is None:
        steady_states = N.dot(samples) + extra_2.T
    else:
        extra_1 = np.full((samples.shape[1], samples.shape[0]), T_shift)
        steady_states = N.dot(T.dot(samples) + extra_1.T) + extra_2.T

    return steady_states


def get_matrices_of_low_dim_polytope(S, lb, ub, min_fluxes, max_fluxes):
    """A Python function to derive the matrices A, Aeq and the vectors b, beq of the low dimensional polytope,
    such that A*x <= b and Aeq*x = beq.

    Keyword arguments:
    samples -- an nxN matrix that contains sample points column-wise
    S -- the stoichiometric matrix
    lb -- lower bounds for the fluxes, i.e., a n-dimensional vector
    ub -- upper bounds for the fluxes, i.e., a n-dimensional vector
    min_fluxes -- minimum values of the fluxes, i.e., a n-dimensional vector
    max_fluxes -- maximum values for the fluxes, i.e., a n-dimensional vector
    """

    n = S.shape[1]
    m = S.shape[0]
    beq = np.zeros(m)
    Aeq = S

    A = np.zeros((2 * n, n), dtype="float")
    A[0:n] = np.eye(n)
    A[n:] -= np.eye(n, n, dtype="float")

    b = np.concatenate((ub, -lb), axis=0)
    b = np.asarray(b, dtype="float")
    b = np.ascontiguousarray(b, dtype="float")

    for i in range(n):

        width = abs(max_fluxes[i] - min_fluxes[i])

        # Check whether we keep or not the equality
        if width < 1e-07:
            Aeq = np.vstack(
                (
                    Aeq,
                    A[
                        i,
                    ],
                )
            )
            beq = np.append(beq, min(max_fluxes[i], min_fluxes[i]))

    return A, b, Aeq, beq


def get_matrices_of_full_dim_polytope(A, b, Aeq, beq):
    """A Python function to derive the matrix A and the vector b of the full dimensional polytope,
    such that Ax <= b given a low dimensional polytope.

    Keyword arguments:
    A -- an mxn matrix that contains the normal vectors of the facets of the polytope row-wise
    b -- a m-dimensional vector, s.t. A*x <= b
    Aeq -- an kxn matrix that contains the normal vectors of hyperplanes row-wise
    beq -- a k-dimensional vector, s.t. Aeq*x = beq
    """

    nullspace_res = nullspace_sparse(Aeq, beq)
    N = nullspace_res[0]
    N_shift = nullspace_res[1]

    if A.shape[1] != N.shape[0] or N.shape[0] != N_shift.size or N.shape[1] <= 1:
        raise Exception(
            "The computation of the matrix of the right nullspace of the stoichiometric matrix failed."
        )

    product = np.dot(A, N_shift)
    b = np.subtract(b, product)
    A = np.dot(A, N)

    res = remove_almost_redundant_facets(A, b)
    A = res[0]
    b = res[1]

    try:
        res = gmscale(A, 0.99)
        res = apply_scaling(A, b, res[0], res[1])
        A = res[0]
        b = res[1]
        N = np.dot(N, res[2])

        res = remove_almost_redundant_facets(A, b)
        A = res[0]
        b = res[1]
    except:
        print("gmscale failed to compute a good scaling.")

    return A, b, N, N_shift



def correlated_reactions(steady_states, reactions=[], pearson_cutoff = 0.90, indicator_cutoff = 10,
                         cells = 10, cop_coeff = 0.3, lower_triangle = True, verbose = False):
    """A Python function to calculate the pearson correlation matrix of a model
       and filter values based on the copula's indicator

    Keyword arguments:
    steady_states -- A numpy array of the generated steady states fluxes
    reactions -- A list with the model's reactions
    pearson_cutoff -- A cutoff to filter reactions based on pearson coefficient
    indicator_cutoff -- A cutoff to filter reactions based on indicator value
    cells -- Number of cells to compute the copula
    cop_coeff -- A value that narrows or widens the width of the copula's diagonal
    lower_triangle -- A boolean variable that if True plots only the lower triangular matrix
    verbose -- A boolean variable that if True additional information is printed as an output.
    """

    if cop_coeff > 0.4 or cop_coeff < 0.2:
        raise Exception("Input value to cop_coeff parameter must be between 0.2 and 0.4")

    # calculate coefficients to access red and blue copula mass
    cop_coeff_1 = cop_coeff
    cop_coeff_2 = 1 - cop_coeff
    cop_coeff_3 = 1 + cop_coeff

    # compute correlation matrix
    corr_matrix = np.corrcoef(steady_states, rowvar=True)

    # replace not assigned values with 0
    corr_matrix[np.isnan(corr_matrix)] = 0

    # create a copy of correlation matrix to replace/filter values
    filtered_corr_matrix = corr_matrix.copy()

    # find indices of correlation matrix where correlation does not occur
    no_corr_indices = np.argwhere((filtered_corr_matrix < pearson_cutoff) & (filtered_corr_matrix > -pearson_cutoff))

    # replace values from the correlation matrix that do not overcome
    # the pearson cutoff with 0
    for i in range(0, no_corr_indices.shape[0]):
        index1 = no_corr_indices[i][0]
        index2 = no_corr_indices[i][1]

        filtered_corr_matrix[index1, index2] = 0

    # if user does not provide an indicator cutoff then do not proceed
    # with the filtering of the correlation matrix
    if indicator_cutoff == 0:
        if lower_triangle == True:
            filtered_corr_matrix[np.triu_indices(filtered_corr_matrix.shape[0], 1)] = np.nan
            np.fill_diagonal(filtered_corr_matrix, 1)
            return filtered_corr_matrix
        else:
            np.fill_diagonal(filtered_corr_matrix, 1)
            return filtered_corr_matrix
    else:
        # a dictionary that will store for each filtered reaction combination,
        # the pearson correlation value, the copula's indicator value
        # and the correlation classification
        indicator_dict = {}

        # keep only the lower triangle
        corr_matrix = np.tril(corr_matrix)
        # replace diagonal values with 0
        np.fill_diagonal(corr_matrix, 0)

        # find indices of correlation matrix where correlation occurs
        corr_indices = np.argwhere((corr_matrix > pearson_cutoff) | (corr_matrix < -pearson_cutoff))

        # compute copula for each set of correlated reactions
        for i in range(0, corr_indices.shape[0]):

            index1 = corr_indices[i][0]
            index2 = corr_indices[i][1]

            reaction1 = reactions[index1]
            reaction2 = reactions[index2]

            flux1 = steady_states[index1]
            flux2 = steady_states[index2]

            copula = compute_copula(flux1, flux2, cells)
            rows, cols = copula.shape

            red_mass = 0
            blue_mass = 0
            indicator = 0

            for row in range(rows):
                for col in range(cols):
                    # values in the diagonal
                    if ((row-col >= -cop_coeff_1*rows) & (row-col <= cop_coeff_1*rows)):
                        # values near the top left and bottom right corner
                        if ((row+col < cop_coeff_2*rows) | (row+col > cop_coeff_3*rows)):
                            red_mass = red_mass + copula[row][col]
                    else:
                        # values near the top right and bottom left corner
                        if ((row+col >= cop_coeff_2*rows-1) & (row+col <= cop_coeff_3*rows-1)):
                            blue_mass = blue_mass + copula[row][col]

            indicator = (red_mass+1e-9) / (blue_mass+1e-9)

            # classify specific pair of reactions as positive or negative
            # correlated based on indicator cutoff
            if indicator > indicator_cutoff:
                pearson = filtered_corr_matrix[index1, index2]
                indicator_dict[reaction1 + "~" + reaction2] = {'pearson': pearson,
                                                               'indicator': indicator,
                                                               'classification': "positive"}

            elif indicator < 1/indicator_cutoff:
                pearson = filtered_corr_matrix[index1, index2]
                indicator_dict[reaction1 + "~" + reaction2] = {'pearson': pearson,
                                                               'indicator': indicator,
                                                               'classification': "negative"}

            # if they do not overcome the cutoff replace their corresponding
            # value in the correlation matrix with 0
            else:
                filtered_corr_matrix[index1, index2] = 0
                filtered_corr_matrix[index2, index1] = 0
                pearson = filtered_corr_matrix[index1, index2]
                indicator_dict[reaction1 + "~" + reaction2] = {'pearson': pearson,
                                                               'indicator': indicator,
                                                               'classification': "no correlation"}

            if verbose == True:
                print("Completed process of",i+1,"from",corr_indices.shape[0],"copulas")

        if lower_triangle == True:
            filtered_corr_matrix[np.triu_indices(filtered_corr_matrix.shape[0], 1)] = np.nan
            np.fill_diagonal(filtered_corr_matrix, 1)
            return filtered_corr_matrix, indicator_dict

        else:
            np.fill_diagonal(filtered_corr_matrix, 1)
            return filtered_corr_matrix, indicator_dict



def cluster_corr_reactions(correlation_matrix, reactions, linkage="ward",
                           t = 4.0, correction=True):
    """A Python function for hierarchical clustering of the correlation matrix

    Keyword arguments:
    correlation_matrix -- A numpy 2D array of a correlation matrix
    reactions -- A list with the model's reactions
    linkage -- linkage defines the type of linkage.
               Available linkage types are: single, average, complete, ward.
    t -- A threshold that defines a threshold that cuts the dendrogram
         at a specific height and produces clusters
    correction -- A boolean variable that if True converts the values of the
                  the correlation matrix to absolute values.
    """

    # function to return a nested list with grouped reactions based on clustering
    def clusters_list(reactions, labels):
        clusters = []
        unique_labels = np.unique(labels)
        for label in unique_labels:
            cluster = []
            label_where = np.where(labels == label)[0]
            for where in label_where:
                cluster.append(reactions[where])
            clusters.append(cluster)
        return clusters

    if correction == True:
        dissimilarity_matrix = 1 - abs(correlation_matrix)
    else:
        dissimilarity_matrix = 1 - correlation_matrix

    Z = hierarchy.linkage(dissimilarity_matrix, linkage)
    labels = hierarchy.fcluster(Z, t, criterion='distance')

    clusters = clusters_list(reactions, labels)
    return dissimilarity_matrix, labels, clusters



def graph_corr_matrix(correlation_matrix, reactions, correction=True,
                      clusters=[], subgraph_nodes = 5):
    """A Python function that creates the main graph and its subgraphs
       from a correlation matrix.

    Keyword arguments:
    correlation_matrix -- A numpy 2D array of a correlation matrix.
    reactions -- A list with the model's reactions.
    correction -- A boolean variable that if True converts the values of the
                  the correlation matrix to absolute values.
    clusters -- A nested list with clustered reactions created from the "" function.
    subgraph_nodes -- A variable that specifies a cutoff for a graph's nodes.
                      It filters subgraphs with low number of nodes..
    """

    graph_matrix = correlation_matrix.copy()
    np.fill_diagonal(graph_matrix, 0)

    if correction == True:
        graph_matrix = abs(graph_matrix)

    G = nx.from_numpy_array(graph_matrix)
    G = nx.relabel_nodes(G, lambda x: reactions[x])

    pos = nx.spring_layout(G)
    unconnected_nodes = list(nx.isolates(G))
    G.remove_nodes_from(unconnected_nodes)
    G_nodes = G.nodes()

    graph_list = []
    layout_list = []

    graph_list.append(G)
    layout_list.append(pos)

    subgraphs = [G.subgraph(c) for c in connected_components(G)]
    H_nodes_list = []

    for i in range(len(subgraphs)):
        if len(subgraphs[i].nodes()) > subgraph_nodes and len(subgraphs[i].nodes()) != len(G_nodes):
            H = G.subgraph(subgraphs[i].nodes())
            for cluster in clusters:
                if H.has_node(cluster[0]) and H.nodes() not in H_nodes_list:
                    H_nodes_list.append(H.nodes())

                    pos = nx.spring_layout(H)
                    graph_list.append(H)
                    layout_list.append(pos)

    return graph_list, layout_list



# dingo_walk : a python library for metabolic networks sampling and analysis
# dingo_walk is part of GeomScale project

# Copyright (c) 2021 Apostolos Chalkis

# Licensed under GNU LGPL.3, see LICENCE file

import numpy as np
import sys
import os
import pickle
from dingo_walk.loading_models import read_json_file
from dingo_walk.nullspace import nullspace_dense, nullspace_sparse
from dingo_walk.scaling import gmscale
from dingo_walk.utils import (
    apply_scaling,
    remove_almost_redundant_facets,
    map_samples_to_steady_states,
    get_matrices_of_low_dim_polytope,
    get_matrices_of_full_dim_polytope,
)
from dingo_walk.illustrations import (
    plot_copula,
    plot_histogram,
)
from dingo_walk.parser import dingo_walk_args
from dingo_walk.MetabolicNetwork import MetabolicNetwork
from dingo_walk.PolytopeSampler import PolytopeSampler

from dingo_walk.pyoptinterface_based_impl import fba, fva, inner_ball, remove_redundant_facets, set_default_solver

from volestipy import HPolytope


def get_name(args_network):

    position = [pos for pos, char in enumerate(args_network) if char == "/"]

    if args_network[-4:] == "json":
        if position == []:
            name = args_network[0:-5]
        else:
            name = args_network[(position[-1] + 1) : -5]
    elif args_network[-3:] == "mat":
        if position == []:
            name = args_network[0:-4]
        else:
            name = args_network[(position[-1] + 1) : -4]

    return name


def dingo_walk_main():
    """A function that (a) reads the inputs using argparse package, (b) calls the proper dingo_walk pipeline
    and (c) saves the outputs using pickle package
    """

    args = dingo_walk_args()

    if args.metabolic_network is None and args.polytope is None and not args.histogram:
        raise Exception(
            "You have to give as input either a model or a polytope derived from a model."
        )

    if args.metabolic_network is None and ((args.fva) or (args.fba)):
        raise Exception("You have to give as input a model to apply FVA or FBA method.")

    if args.output_directory == None:
        output_path_dir = os.getcwd()
    else:
        output_path_dir = args.output_directory

    if os.path.isdir(output_path_dir) == False:
        os.mkdir(output_path_dir)

    # Move to the output directory
    os.chdir(output_path_dir)

    set_default_solver(args.solver)

    if args.model_name is None:
        if args.metabolic_network is not None:
            name = get_name(args.metabolic_network)
    else:
        name = args.model_name

    if args.histogram:

        if args.steady_states is None:
            raise Exception(
                "A path to a pickle file that contains steady states of the model has to be given."
            )

        if args.metabolites_reactions is None:
            raise Exception(
                "A path to a pickle file that contains the names of the metabolites and the reactions of the model has to be given."
            )

        if int(args.reaction_index) <= 0:
            raise Exception("The index of the reaction has to be a positive integer.")

        file = open(args.steady_states, "rb")
        steady_states = pickle.load(file)
        file.close()

        file = open(args.metabolites_reactions, "rb")
        model = pickle.load(file)
        file.close()

        reactions = model.reactions

        if int(args.reaction_index) > len(reactions):
            raise Exception(
                "The index of the reaction has not to be exceed the number of reactions."
            )

        if int(args.n_bins) <= 0:
            raise Exception("The number of bins has to be a positive integer.")

        plot_histogram(
            steady_states[int(args.reaction_index) - 1],
            reactions[int(args.reaction_index) - 1],
            int(args.n_bins),
        )

    elif args.fva:

        if args.metabolic_network[-4:] == "json":
            model = MetabolicNetwork.fom_json(args.metabolic_network)
        elif args.metabolic_network[-3:] == "mat":
            model = MetabolicNetwork.fom_mat(args.metabolic_network)
        else:
            raise Exception("An unknown format file given.")

        model.set_solver(args.solver)

        result_obj = model.fva()

        with open("dingo_walk_fva_" + name + ".pckl", "wb") as dingo_walk_fva_file:
            pickle.dump(result_obj, dingo_walk_fva_file)

    elif args.fba:

        if args.metabolic_network[-4:] == "json":
            model = MetabolicNetwork.fom_json(args.metabolic_network)
        elif args.metabolic_network[-3:] == "mat":
            model = MetabolicNetwork.fom_mat(args.metabolic_network)
        else:
            raise Exception("An unknown format file given.")

        model.set_solver(args.solver)

        result_obj = model.fba()

        with open("dingo_walk_fba_" + name + ".pckl", "wb") as dingo_walk_fba_file:
            pickle.dump(result_obj, dingo_walk_fba_file)

    elif args.metabolic_network is not None:

        if args.metabolic_network[-4:] == "json":
            model = MetabolicNetwork.fom_json(args.metabolic_network)
        elif args.metabolic_network[-3:] == "mat":
            model = MetabolicNetwork.fom_mat(args.metabolic_network)
        else:
            raise Exception("An unknown format file given.")

        sampler = PolytopeSampler(model)

        if args.preprocess_only:

            sampler.get_polytope()

            polytope_info = (
                sampler,
                name,
            )

            with open("dingo_walk_model_" + name + ".pckl", "wb") as dingo_walk_model_file:
                pickle.dump(model, dingo_walk_model_file)

            with open(
                "dingo_walk_polytope_sampler_" + name + ".pckl", "wb"
            ) as dingo_walk_polytope_file:
                pickle.dump(polytope_info, dingo_walk_polytope_file)

        else:

            steady_states = sampler.generate_steady_states(
                int(args.effective_sample_size),
                args.psrf_check,
                args.parallel_mmcs,
                int(args.num_threads),
            )

            polytope_info = (
                sampler,
                name,
            )

            with open("dingo_walk_model_" + name + ".pckl", "wb") as dingo_walk_model_file:
                pickle.dump(model, dingo_walk_model_file)

            with open(
                "dingo_walk_polytope_sampler_" + name + ".pckl", "wb"
            ) as dingo_walk_polytope_file:
                pickle.dump(polytope_info, dingo_walk_polytope_file)

            with open(
                "dingo_walk_steady_states_" + name + ".pckl", "wb"
            ) as dingo_walk_steadystates_file:
                pickle.dump(steady_states, dingo_walk_steadystates_file)

    else:

        file = open(args.polytope, "rb")
        input_obj = pickle.load(file)
        file.close()
        sampler = input_obj[0]

        if isinstance(sampler, PolytopeSampler):

            steady_states = sampler.generate_steady_states(
                int(args.effective_sample_size),
                args.psrf_check,
                args.parallel_mmcs,
                int(args.num_threads),
            )

        else:
            raise Exception("The input file has to be generated by dingo_walk package.")

        if args.model_name is None:
            name = input_obj[-1]

        polytope_info = (
            sampler,
            name,
        )

        with open(
            "dingo_walk_polytope_sampler" + name + "_improved.pckl", "wb"
        ) as dingo_walk_polytope_file:
            pickle.dump(polytope_info, dingo_walk_polytope_file)

        with open("dingo_walk_steady_states_" + name + ".pckl", "wb") as dingo_walk_network_file:
            pickle.dump(steady_states, dingo_walk_network_file)


if __name__ == "__main__":

    dingo_walk_main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dingo_walk']

package_data = \
{'': ['*'], 'dingo_walk': ['bindings/*']}

install_requires = \
['Cython>=0.29.22,<0.30.0',
 'argparse>=1.4.0,<2.0.0',
 'cobra>=0.29.1,<0.30.0',
 'kaleido==0.2.1',
 'matplotlib>=3.4.1,<4.0.0',
 'networkx==3.1',
 'plotly>=5.11.0,<6.0.0',
 'pyoptinterface[highs]>=0.2.7,<0.3.0',
 'scipy>=1.6.1,<2.0.0',
 'simplejson>=3.17.2,<4.0.0',
 'sparseqr>=1.2.1,<2.0.0']

setup_kwargs = {
    'name': 'dingo-walk',
    'version': '0.1.5',
    'description': 'High dimensional polytope sampling in Python with a set of tools for metabolic network sampling and analysis.',
    'long_description': '<p align="center"><img src="https://raw.githubusercontent.com/GeomScale/dingo/refs/heads/develop/doc/logo/dingo.jpg" width="260" height="260"></p>\n\n**dingo_walk** is a Python package that analyzes metabolic networks.\nIt relies on high dimensional sampling with Markov Chain Monte Carlo (MCMC)\nmethods and fast optimization methods to analyze the possible states of a\nmetabolic network. To perform MCMC sampling, `dingo_walk` relies on the `C++` library\n[volesti](https://github.com/GeomScale/volume_approximation), which provides\nseveral algorithms for sampling convex polytopes.\n`dingo_walk` also performs two standard methods to analyze the flux space of a\nmetabolic network, namely Flux Balance Analysis and Flux Variability Analysis.\n\n`dingo_walk` is part of [GeomScale](https://geomscale.github.io/) project.\n\n[![unit-tests](https://github.com/GeomScale/dingo/workflows/dingo_walk-ubuntu/badge.svg)](https://github.com/GeomScale/dingo/actions?query=workflow%3Adingo_walk-ubuntu)\n[![Tutorial In Colab](https://colab.research.google.com/assets/colab-badge.svg)](https://colab.research.google.com/github/GeomScale/dingo/blob/develop/tutorials/dingo_tutorial.ipynb)\n[![Chat](https://badges.gitter.im/geomscale.png)](https://gitter.im/GeomScale/community?utm_source=share-link&utm_medium=link&utm_campaign=share-link)\n\n\n## Installation\n\n**Note:** Python version should be 3.8.x. You can check this by running the following command in your terminal:\n```bash\npython --version\n```\nIf you have a different version of Python installed, you\'ll need to install it ([start here](https://linuxize.com/post/how-to-install-python-3-8-on-ubuntu-18-04/)) and update-alternatives ([start here](https://linuxhint.com/update_alternatives_ubuntu/))\n\n**Note:** If you are using `GitHub Codespaces`. Start [here](https://docs.github.com/en/codespaces/setting-up-your-project-for-codespaces/adding-a-dev-container-configuration/setting-up-your-python-project-for-codespaces) to set the python version. Once your Python version is `3.8.x` you can start following the below instructions.\n\n\n\nTo load the submodules that dingo_walk uses, run\n\n````bash\ngit submodule update --init\n````\n\nYou will need to download and unzip the Boost library:\n```\nwget -O boost_1_76_0.tar.bz2 https://archives.boost.io/release/1.76.0/source/boost_1_76_0.tar.bz2\ntar xjf boost_1_76_0.tar.bz2\nrm boost_1_76_0.tar.bz2\n```\n\nYou will also need to download and unzip the lpsolve library:\n```\nwget https://sourceforge.net/projects/lpsolve/files/lpsolve/5.5.2.11/lp_solve_5.5.2.11_source.tar.gz\ntar xzvf lp_solve_5.5.2.11_source.tar.gz\nrm lp_solve_5.5.2.11_source.tar.gz\n```\n\nThen, you need to install the dependencies for the PySPQR library; for Debian/Ubuntu Linux, run\n\n```bash\nsudo apt-get update -y\nsudo apt-get install -y libsuitesparse-dev\n```\n\nTo install the Python dependencies, `dingo_walk` is using [Poetry](https://python-poetry.org/),\n```\ncurl -sSL https://install.python-poetry.org | python3 - --version 1.3.2\npoetry shell\npoetry install\n```\n\nYou can install the [Gurobi solver](https://www.gurobi.com/) for faster linear programming optimization. Run\n\n```\npip3 install -i https://pypi.gurobi.com gurobipy\n```\n\nThen, you will need a [license](https://www.gurobi.com/downloads/end-user-license-agreement-academic/). For more information, we refer to the Gurobi [download center](https://www.gurobi.com/downloads/).\n\n\n\n\n## Unit tests\n\nNow, you can run the unit tests by the following commands (with the default solver `highs`):\n```\npython3 tests/fba.py\npython3 tests/full_dimensional.py\npython3 tests/max_ball.py\npython3 tests/scaling.py\npython3 tests/rounding.py\npython3 tests/sampling.py\n```\n\nIf you have installed Gurobi successfully, then run\n```\npython3 tests/fba.py gurobi\npython3 tests/full_dimensional.py gurobi\npython3 tests/max_ball.py gurobi\npython3 tests/scaling.py gurobi\npython3 tests/rounding.py gurobi\npython3 tests/sampling.py gurobi\n```\n\n## Tutorial\n\nYou can have a look at our [Google Colab notebook](https://colab.research.google.com/github/GeomScale/dingo/blob/develop/tutorials/dingo_tutorial.ipynb)\non how to use `dingo_walk`.\n\n\n## Documentation\n\n\nIt quite simple to use dingo_walk in your code. In general, dingo_walk provides two classes:\n\n- `metabolic_network` represents a metabolic network\n- `polytope_sampler` can be used to sample from the flux space of a metabolic network or from a general convex polytope.\n\n The following script shows how you could sample steady states of a metabolic network with dingo_walk. To initialize a metabolic network object you have to provide the path to the `json` file as those in [BiGG](http://bigg.ucsd.edu/models) dataset or the `mat` file (using the `matlab` wrapper in folder `/ext_data` to modify a standard `mat` file of a model as those in BiGG dataset):\n\n```python\nfrom dingo_walk import MetabolicNetwork, PolytopeSampler\n\nmodel = MetabolicNetwork.from_json(\'path/to/model_file.json\')\nsampler = PolytopeSampler(model)\nsteady_states = sampler.generate_steady_states()\n```\n\n`dingo_walk` can also load a model given in `.sbml` format using the following command,\n\n```python\nmodel = MetabolicNetwork.from_sbml(\'path/to/model_file.sbml\')\n```\n\nThe output variable `steady_states` is a `numpy` array that contains the steady states of the model column-wise. You could ask from the `sampler` for more statistical guarantees on sampling,\n\n```python\nsteady_states = sampler.generate_steady_states(ess=2000, psrf = True)\n```\n\nThe `ess` stands for the effective sample size (ESS) (default value is `1000`) and the `psrf` is a flag to request an upper bound equal to 1.1 for the value of the  *potential scale reduction factor* of each marginal flux (default option is `False`).\n\nYou could also ask for parallel MMCS algorithm,\n\n```python\nsteady_states = sampler.generate_steady_states(ess=2000, psrf = True,\n                                               parallel_mmcs = True, num_threads = 2)\n```\n\nThe default option is to run the sequential [Multiphase Monte Carlo Sampling algorithm](https://arxiv.org/abs/2012.05503) (MMCS) algorithm.\n\n**Tip**: After the first run of MMCS algorithm the polytope stored in object `sampler` is usually more rounded than the initial one. Thus, the function `generate_steady_states()` becomes more efficient from run to run.\n\n\n#### Rounding the polytope\n\n`dingo_walk` provides three methods to round a polytope: (i) Bring the polytope to John position by apllying to it the transformation that maps the largest inscribed ellipsoid of the polytope to the unit ball, (ii) Bring the polytope to near-isotropic position by using uniform sampling with Billiard Walk, (iii) Apply to the polytope the transformation that maps the smallest enclosing ellipsoid of a uniform sample from the interior of the polytope to the unit ball.\n\n```python\nfrom dingo_walk import MetabolicNetwork, PolytopeSampler\n\nmodel = MetabolicNetwork.from_json(\'path/to/model_file.json\')\nsampler = PolytopeSampler(model)\nA, b, N, N_shift = sampler.get_polytope()\n\nA_rounded, b_rounded, Tr, Tr_shift = sampler.round_polytope(A, b, method="john_position")\nA_rounded, b_rounded, Tr, Tr_shift = sampler.round_polytope(A, b, method="isotropic_position")\nA_rounded, b_rounded, Tr, Tr_shift = sampler.round_polytope(A, b, method="min_ellipsoid")\n```\n\nThen, to sample from the rounded polytope, the user has to call the following static method of PolytopeSampler class,\n\n```python\nsamples = sample_from_polytope(A_rounded, b_rounded)\n```\n\nLast you can map the samples back to steady states,\n\n```python\nfrom dingo_walk import map_samples_to_steady_states\n\nsteady_states = map_samples_to_steady_states(samples, N, N_shift, Tr, Tr_shift)\n```\n\n#### Other MCMC sampling methods\n\nTo use any other MCMC sampling method that `dingo_walk` provides you can use the following piece of code:\n\n```python\nsampler = polytope_sampler(model)\nsteady_states = sampler.generate_steady_states_no_multiphase() #default parameters (method = \'billiard_walk\', n=1000, burn_in=0, thinning=1)\n```\n\nThe MCMC methods that dingo_walk (through `volesti` library) provides are the following: (i) \'cdhr\': Coordinate Directions Hit-and-Run, (ii) \'rdhr\': Random Directions Hit-and-Run,\n(iii) \'billiard_walk\', (iv) \'ball_walk\', (v) \'dikin_walk\', (vi) \'john_walk\', (vii) \'vaidya_walk\'.\n\n\n\n#### Switch the linear programming solver\n\nWe use `pyoptinterface` to interface with the linear programming solvers. To switch the solver that `dingo_walk` uses, you can use the `set_default_solver` function. The default solver is `highs` and you can switch to `gurobi` by running,\n\n```python\nfrom dingo_walk import set_default_solver\nset_default_solver("gurobi")\n```\n\nYou can also switch to other solvers that `pyoptinterface` supports, but we recommend using `highs` or `gurobi`. If you have issues with the solver, you can check the `pyoptinterface` [documentation](https://metab0t.github.io/PyOptInterface/getting_started.html).\n\n### Apply FBA and FVA methods\n\nTo apply FVA and FBA methods you have to use the class `metabolic_network`,\n\n```python\nfrom dingo_walk import MetabolicNetwork\n\nmodel = MetabolicNetwork.from_json(\'path/to/model_file.json\')\nfva_output = model.fva()\n\nmin_fluxes = fva_output[0]\nmax_fluxes = fva_output[1]\nmax_biomass_flux_vector = fva_output[2]\nmax_biomass_objective = fva_output[3]\n```\n\nThe output of FVA method is tuple that contains `numpy` arrays. The vectors `min_fluxes` and `max_fluxes` contains the minimum and the maximum values of each flux. The vector `max_biomass_flux_vector` is the optimal flux vector according to the biomass objective function and `max_biomass_objective` is the value of that optimal solution.\n\nTo apply FBA method,\n\n```python\nfba_output = model.fba()\n\nmax_biomass_flux_vector = fba_output[0]\nmax_biomass_objective = fba_output[1]\n```\n\nwhile the output vectors are the same with the previous example.\n\n\n\n### Set the restriction in the flux space\n\nFVA and FBA,  restrict the flux space to the set of flux vectors that have an objective value equal to the optimal value of the function. dingo_walk allows for a more  relaxed option where you could ask for flux vectors that have an objective value equal to at least a percentage of the optimal value,\n\n```python\nmodel.set_opt_percentage(90)\nfva_output = model.fva()\n\n# the same restriction in the flux space holds for the sampler\nsampler = polytope_sampler(model)\nsteady_states = sampler.generate_steady_states()\n```\n\nThe default percentage is `100%`.\n\n\n\n### Change the objective function\n\nYou could also set an alternative objective function. For example, to maximize the 1st reaction of the model,\n\n```python\nn = model.num_of_reactions()\nobj_fun = np.zeros(n)\nobj_fun[0] = 1\nmodel.objective_function(obj_fun)\n\n# apply FVA using the new objective function\nfva_output = model.fva()\n# sample from the flux space by restricting\n# the fluxes according to the new objective function\nsampler = polytope_sampler(model)\nsteady_states = sampler.generate_steady_states()\n```\n\n\n\n### Plot flux marginals\n\nThe generated steady states can be used to estimate the marginal density function of each flux. You can plot the histogram using the samples,\n\n```python\nfrom dingo_walk import plot_histogram\n\nmodel = MetabolicNetwork.from_json(\'path/to/e_coli_core.json\')\nsampler = PolytopeSampler(model)\nsteady_states = sampler.generate_steady_states(ess = 3000)\n\n# plot the histogram for the 14th reaction in e-coli (ACONTa)\nreactions = model.reactions\nplot_histogram(\n        steady_states[13],\n        reactions[13],\n        n_bins = 60,\n        )\n```\n\nThe default number of bins is 60. dingo_walk uses the package `matplotlib` for plotting.\n\n![histogram](https://raw.githubusercontent.com/GeomScale/dingo/refs/heads/develop/doc/e_coli_aconta.png)\n\n### Plot a copula between two fluxes\n\nThe generated steady states can be used to estimate and plot the copula between two fluxes. You can plot the copula using the samples,\n\n```python\nfrom dingo_walk import plot_copula\n\nmodel = MetabolicNetwork.from_json(\'path/to/e_coli_core.json\')\nsampler = PolytopeSampler(model)\nsteady_states = sampler.generate_steady_states(ess = 3000)\n\n# plot the copula between the 13th (PPC) and the 14th (ACONTa) reaction in e-coli\nreactions = model.reactions\n\ndata_flux2=[steady_states[12],reactions[12]]\ndata_flux1=[steady_states[13],reactions[13]]\n\nplot_copula(data_flux1, data_flux2, n=10)\n```\n\nThe default number of cells is 5x5=25. dingo_walk uses the package `plotly` for plotting.\n\n![histogram](https://raw.githubusercontent.com/GeomScale/dingo/refs/heads/develop/doc/aconta_ppc_copula.png)\n\n\n',
    'author': 'Apostolos Chalkis',
    'author_email': 'tolis.chal@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



# dingo_walk : a python library for metabolic networks sampling and analysis
# dingo_walk is part of GeomScale project

# Copyright (c) 2021 Apostolos Chalkis
# Copyright (c) 2021 Vissarion Fisikopoulos
# Copyright (c) 2024 Ke Shi

# Licensed under GNU LGPL.3, see LICENCE file

import numpy as np
import sys
from typing import Dict
import cobra
from dingo_walk.loading_models import read_json_file, read_mat_file, read_sbml_file, parse_cobra_model
from dingo_walk.pyoptinterface_based_impl import fba,fva,inner_ball,remove_redundant_facets

class MetabolicNetwork:
    def __init__(self, tuple_args):

        self._parameters = {}
        self._parameters["opt_percentage"] = 100
        self._parameters["distribution"] = "uniform"
        self._parameters["nullspace_method"] = "sparseQR"
        self._parameters["solver"] = None

        if len(tuple_args) != 10:
            raise Exception(
                "An unknown input format given to initialize a metabolic network object."
            )

        self._lb = tuple_args[0]
        self._ub = tuple_args[1]
        self._S = tuple_args[2]
        self._metabolites = tuple_args[3]
        self._reactions = tuple_args[4]
        self._biomass_index = tuple_args[5]
        self._objective_function = tuple_args[6]
        self._medium = tuple_args[7]
        self._medium_indices = tuple_args[8]
        self._exchanges = tuple_args[9]

        try:
            if self._biomass_index is not None and (
                self._lb.size != self._ub.size
                or self._lb.size != self._S.shape[1]
                or len(self._metabolites) != self._S.shape[0]
                or len(self._reactions) != self._S.shape[1]
                or self._objective_function.size != self._S.shape[1]
                or (self._biomass_index < 0)
                or (self._biomass_index > self._objective_function.size)
            ):
                raise Exception(
                    "Wrong tuple format given to initialize a metabolic network object."
                )
        except LookupError as error:
            raise error.with_traceback(sys.exc_info()[2])

    @classmethod
    def from_json(cls, arg):
        if (not isinstance(arg, str)) or (arg[-4:] != "json"):
            raise Exception(
                "An unknown input format given to initialize a metabolic network object."
            )

        return cls(read_json_file(arg))

    @classmethod
    def from_mat(cls, arg):
        if (not isinstance(arg, str)) or (arg[-3:] != "mat"):
            raise Exception(
                "An unknown input format given to initialize a metabolic network object."
            )

        return cls(read_mat_file(arg))

    @classmethod
    def from_sbml(cls, arg):
        if (not isinstance(arg, str)) and ((arg[-3:] == "xml") or (arg[-4] == "sbml")):
            raise Exception(
                "An unknown input format given to initialize a metabolic network object."
            )

        return cls(read_sbml_file(arg))

    @classmethod
    def from_cobra_model(cls, arg):
        if (not isinstance(arg, cobra.core.model.Model)):
            raise Exception(
                "An unknown input format given to initialize a metabolic network object."
            )

        return cls(parse_cobra_model(arg))

    def fva(self):
        """A member function to apply the FVA method on the metabolic network."""

        return fva(
            self._lb,
            self._ub,
            self._S,
            self._objective_function,
            self._parameters["opt_percentage"],
            self._parameters["solver"]
        )

    def fba(self):
        """A member function to apply the FBA method on the metabolic network."""
        return fba(self._lb, self._ub, self._S, self._objective_function, self._parameters["solver"])

    @property
    def lb(self):
        return self._lb

    @property
    def ub(self):
        return self._ub

    @property
    def S(self):
        return self._S

    @property
    def metabolites(self):
        return self._metabolites

    @property
    def reactions(self):
        return self._reactions

    @property
    def biomass_index(self):
        return self._biomass_index

    @property
    def objective_function(self):
        return self._objective_function

    @property
    def medium(self):
        return self._medium

    @property
    def exchanges(self):
        return self._exchanges

    @property
    def parameters(self):
        return self._parameters

    @property
    def get_as_tuple(self):
        return (
            self._lb,
            self._ub,
            self._S,
            self._metabolites,
            self._reactions,
            self._biomass_index,
            self._objective_function,
            self._medium,
            self._inter_medium,
            self._exchanges
        )

    def num_of_reactions(self):
        return len(self._reactions)

    def num_of_metabolites(self):
        return len(self._metabolites)

    @lb.setter
    def lb(self, value):
        self._lb = value

    @ub.setter
    def ub(self, value):
        self._ub = value

    @S.setter
    def S(self, value):
        self._S = value

    @metabolites.setter
    def metabolites(self, value):
        self._metabolites = value

    @reactions.setter
    def reactions(self, value):
        self._reactions = value

    @biomass_index.setter
    def biomass_index(self, value):
        self._biomass_index = value

    @objective_function.setter
    def objective_function(self, value):
        self._objective_function = value


    @medium.setter
    def medium(self, medium: Dict[str, float]) -> None:
        """Set the constraints on the model exchanges.

        `model.medium` returns a dictionary of the bounds for each of the
        boundary reactions, in the form of `{rxn_id: rxn_bound}`, where `rxn_bound`
        specifies the absolute value of the bound in direction of metabolite
        creation (i.e., lower_bound for `met <--`, upper_bound for `met -->`)

        Parameters
        ----------
        medium: dict
            The medium to initialize. medium should be a dictionary defining
            `{rxn_id: bound}` pairs.
        """

        def set_active_bound(reaction: str, reac_index: int, bound: float) -> None:
            """Set active bound.

            Parameters
            ----------
            reaction: cobra.Reaction
                Reaction to set
            bound: float
                Value to set bound to. The bound is reversed and set as lower bound
                if reaction has reactants (metabolites that are consumed). If reaction
                has reactants, it seems the upper bound won't be set.
            """
            if any(x < 0 for x in  list(self._S[:, reac_index])):
                self._lb[reac_index] = -bound
            elif any(x > 0 for x in  list(self._S[:, reac_index])):
                self._ub[reac_index] = bound

        # Set the given media bounds
        media_rxns = []
        exchange_rxns = frozenset(self.exchanges)
        for rxn_id, rxn_bound in medium.items():
            if rxn_id not in exchange_rxns:
                logger.warning(
                    f"{rxn_id} does not seem to be an an exchange reaction. "
                    f"Applying bounds anyway."
                )
            media_rxns.append(rxn_id)

            reac_index = self._reactions.index(rxn_id)

            set_active_bound(rxn_id, reac_index, rxn_bound)

        frozen_media_rxns = frozenset(media_rxns)

        # Turn off reactions not present in media
        for rxn_id in exchange_rxns - frozen_media_rxns:
            """
            is_export for us, needs to check on the S
            order reactions to their lb and ub
            """
            # is_export = rxn.reactants and not rxn.products
            reac_index = self._reactions.index(rxn_id)
            products = np.any(self._S[:,reac_index] > 0)
            reactants_exist = np.any(self._S[:,reac_index] < 0)
            is_export = True if not products and reactants_exist else False
            set_active_bound(
                rxn_id, reac_index, min(0.0, -self._lb[reac_index] if is_export else self._ub[reac_index])
            )

    def set_solver(self, solver: str):
        self._parameters["solver"] = solver

    def set_nullspace_method(self, value):

        self._parameters["nullspace_method"] = value

    def set_opt_percentage(self, value):

        self._parameters["opt_percentage"] = value

    def shut_down_reaction(self, index_val):

        if (
            (not isinstance(index_val, int))
            or index_val < 0
            or index_val >= self._S.shape[1]
        ):
            raise Exception("The input does not correspond to a proper reaction index.")

        self._lb[index_val] = 0
        self._ub[index_val] = 0
